import numpy as np
import xarray as xr
from struct import unpack
import warnings
from pathlib import Path
import logging

from . import nortek_defs
from .base import _find_userdata, _create_dataset, _handle_nan, _abspath
from .. import time
from datetime import datetime
from ..tools import misc as tbx
from ..rotate.vector import _calc_omat
from ..rotate.base import _set_coords
from ..rotate import api as rot


def read_nortek(filename, userdata=True, debug=False, do_checksum=False,
                nens=None, **kwargs):
    """Read a classic Nortek (AWAC and Vector) datafile

    Parameters
    ----------
    filename : string
        Filename of Nortek file to read.
    userdata : True, False, or string of userdata.json filename
        (default ``True``) Whether to read the '<base-filename>.userdata.json'
        file.
    do_checksum : bool (default False)
        Whether to perform the checksum of each data block.
    nens : None (default: read entire file), int, or 2-element tuple (start, stop)
        Number of pings to read from the file

    Returns
    -------
    ds : xarray.Dataset
        An xarray dataset from the binary instrument data

    """
    # Start debugger logging
    if debug:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
        filepath = Path(filename)
        logfile = filepath.with_suffix('.log')
        logging.basicConfig(filename=str(logfile),
                            filemode='w',
                            level=logging.NOTSET,
                            format='%(name)s - %(levelname)s - %(message)s')

    userdata = _find_userdata(filename, userdata)

    with _NortekReader(filename, debug=debug, do_checksum=do_checksum,
                       nens=nens) as rdr:
        rdr.readfile()
    rdr.dat2sci()
    dat = rdr.data

    # Remove trailing nan's in time and orientation data
    dat = _handle_nan(dat)

    # Search for missing timestamps and interpolate them
    coords = dat['coords']
    t_list = [t for t in coords if 'time' in t]
    for ky in t_list:
        tdat = coords[ky]
        tdat[tdat == 0] = np.NaN
        if np.isnan(tdat).any():
            tag = ky.lstrip('time')
            warnings.warn("Zero/NaN values found in '{}'. Interpolating and "
                          "extrapolating them. To identify which values were filled later, "
                          "look for 0 values in 'status{}'".format(ky, tag))
            tdat = time._fill_time_gaps(
                tdat, sample_rate_hz=dat['attrs']['fs'])
        coords[ky] = tdat

    # Apply rotation matrix and declination
    rotmat = None
    declin = None
    for nm in userdata:
        if 'rotmat' in nm:
            rotmat = userdata[nm]
        elif 'dec' in nm:
            declin = userdata[nm]
        else:
            dat['attrs'][nm] = userdata[nm]

    # Create xarray dataset from upper level dictionary
    ds = _create_dataset(dat)
    ds = _set_coords(ds, ref_frame=ds.coord_sys)

    if 'orientmat' not in ds:
        omat = _calc_omat(ds['heading'].values,
                          ds['pitch'].values,
                          ds['roll'].values,
                          ds.get('orientation_down', None))
        ds['orientmat'] = xr.DataArray(omat,
                                       coords={'earth': ['E', 'N', 'U'],
                                               'inst': ['X', 'Y', 'Z'],
                                               'time': ds['time']},
                                       dims=['earth', 'inst', 'time'])
    if rotmat is not None:
        rot.set_inst2head_rotmat(ds, rotmat, inplace=True)
    if declin is not None:
        rot.set_declination(ds, declin, inplace=True)

    ds['time'] = time.epoch2dt64(ds['time']).astype('datetime64[us]')

    # Close handler
    if debug:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
            handler.close()

    return ds


def _bcd2char(cBCD):
    """
    Taken from the Nortek System Integrator
    Manual "Example Program" Chapter.
    """
    cBCD = min(cBCD, 153)
    c = (cBCD & 15)
    c += 10 * (cBCD >> 4)
    return c


def _bitshift8(val):
    return val >> 8


def _int2binarray(val, n):
    out = np.zeros(n, dtype='bool')
    for idx, n in enumerate(range(n)):
        out[idx] = val & (2 ** n)
    return out


class _NortekReader():
    """A class for reading reading nortek binary files.
    This reader currently only supports AWAC and Vector data formats.

    Parameters
    ----------
    fname : string
        Nortek file filename to read.
    endian : {'<','>'} (optional)
        Specifies if the file is in 'little' or 'big' endian format. By
        default the reader will attempt to determine this.
    debug : {True, False*} (optional)
        Print debug/progress information?
    do_checksum : {True*, False} (optional)
        Specifies whether to perform the checksum.
    bufsize : int (default 100000)
        The size of the read buffer to use.
    nens : None (default: None, read all files), int, or 2-element tuple (start, stop).
        The number of pings to read from the file. By default, the entire file
        is read.

    """
    _lastread = [None, None, None, None, None]
    fun_map = {'0x00': 'read_user_cfg',
               '0x04': 'read_head_cfg',
               '0x05': 'read_hw_cfg',
               '0x07': 'read_vec_checkdata',
               '0x10': 'read_vec_data',
               '0x11': 'read_vec_sysdata',
               '0x12': 'read_vec_hdr',
               '0x71': 'read_microstrain',
               '0x20': 'read_awac_profile',
               }

    def __init__(self, fname, endian=None, debug=False,
                 do_checksum=True, bufsize=100000, nens=None):
        self.fname = fname
        self._bufsize = bufsize
        self.f = open(_abspath(fname), 'rb', 1000)
        self.do_checksum = do_checksum
        self.filesize  # initialize the filesize.
        self.debug = debug
        self.c = 0
        self._dtypes = []
        self._n_start = 0
        try:
            len(nens)
        except TypeError:
            # not a tuple, so we assume None or int
            self._npings = nens
        else:
            if len(nens) != 2:
                raise TypeError('nens must be: None (), int, or len 2')
            warnings.warn("A 'start ensemble' is not yet supported "
                          "for the Nortek reader. This function will read "
                          "the entire file, then crop the beginning at "
                          "nens[0].")
            self._npings = nens[1]
            self._n_start = nens[0]
        if endian is None:
            if unpack('<HH', self.read(4)) == (1445, 24):
                endian = '<'
            elif unpack('>HH', self.read(4)) == (1445, 24):
                endian = '>'
            else:
                raise Exception("I/O error: could not determine the "
                                "'endianness' of the file.  Are you sure this is a Nortek "
                                "file?")
        self.endian = endian
        self.f.seek(0, 0)

        # This is the configuration data:
        self.config = {}
        err_msg = ("I/O error: The file does not "
                   "appear to be a Nortek data file.")
        # Read the header:
        if self.read_id() == 5:
            self.read_hw_cfg()
        else:
            raise Exception()
        if self.read_id() == 4:
            self.read_head_cfg()
        else:
            raise Exception(err_msg)
        if self.read_id() == 0:
            self.read_user_cfg()
        else:
            raise Exception(err_msg)
        if self.config['hdw']['serial_number'][0:3].upper() == 'WPR':
            self.config['config_type'] = 'AWAC'
        elif self.config['hdw']['serial_number'][0:3].upper() == 'VEC':
            self.config['config_type'] = 'ADV'
        # Initialize the instrument type:
        self._inst = self.config.pop('config_type')
        # This is the position after reading the 'hardware',
        # 'head', and 'user' configuration.
        pnow = self.pos

        # Run the appropriate initialization routine (e.g. init_ADV).
        getattr(self, 'init_' + self._inst)()
        self.f.close()  # This has a small buffer, so close it.
        # This has a large buffer...
        self.f = open(_abspath(fname), 'rb', bufsize)
        self.close = self.f.close
        if self._npings is not None:
            self.n_samp_guess = self._npings
        self.f.seek(pnow, 0)  # Seek to the previous position.

        da = self.data['attrs']
        if self.config['n_burst'] > 0:
            fs = round(self.config['fs'], 7)
            da['duty_cycle_n_burst'] = self.config['n_burst']
            da['duty_cycle_interval'] = self.config['burst_interval']
            if fs > 1:
                burst_seconds = self.config['n_burst']/fs
            else:
                burst_seconds = round(1/fs, 3)
            da['duty_cycle_description'] = "{} second bursts collected at {} Hz, with bursts taken every {} minutes".format(
                burst_seconds, fs, self.config['burst_interval']/60)
        self.burst_start = np.zeros(self.n_samp_guess, dtype='bool')
        da['fs'] = self.config['fs']
        da['coord_sys'] = {'XYZ': 'inst',
                           'ENU': 'earth',
                           'beam': 'beam'}[self.config['coord_sys_axes']]
        da['has_imu'] = 0  # Initiate attribute
        if self.debug:
            logging.info('Init completed')

    @property
    def filesize(self,):
        if not hasattr(self, '_filesz'):
            pos = self.pos
            self.f.seek(0, 2)
            # Seek to the end of the file to determine the filesize.
            self._filesz = self.pos
            self.f.seek(pos, 0)  # Return to the initial position.
        return self._filesz

    @property
    def pos(self,):
        return self.f.tell()

    def init_ADV(self,):
        dat = self.data = {'data_vars': {}, 'coords': {}, 'attrs': {},
                           'units': {}, 'sys': {}}
        da = dat['attrs']
        dv = dat['data_vars']
        da['inst_make'] = 'Nortek'
        da['inst_model'] = 'Vector'
        da['inst_type'] = 'ADV'
        da['rotate_vars'] = ['vel']
        dv['beam2inst_orientmat'] = self.config.pop('beam2inst_orientmat')
        self.config['fs'] = 512 / self.config['awac']['avg_interval']
        da.update(self.config['usr'])
        da.update(self.config['adv'])
        da.update(self.config['head'])
        da.update(self.config['hdw'])

        # No apparent way to determine how many samples are in a file
        dlta = self.code_spacing('0x11')
        self.n_samp_guess = int(self.filesize / dlta + 1)
        self.n_samp_guess *= int(self.config['fs'])

    def init_AWAC(self,):
        dat = self.data = {'data_vars': {}, 'coords': {}, 'attrs': {},
                           'units': {}, 'sys': {}}
        da = dat['attrs']
        dv = dat['data_vars']
        da['inst_make'] = 'Nortek'
        da['inst_model'] = 'AWAC'
        da['inst_type'] = 'ADCP'
        dv['beam2inst_orientmat'] = self.config.pop('beam2inst_orientmat')
        da['rotate_vars'] = ['vel']
        self.config['fs'] = 1. / self.config['awac']['avg_interval']
        da.update(self.config['usr'])
        da.update(self.config['awac'])
        da.update(self.config['head'])
        da.update(self.config['hdw'])

        space = self.code_spacing('0x20')
        if space == 0:
            # code spacing is zero if there's only 1 profile
            self.n_samp_guess = 1
        else:
            self.n_samp_guess = int(self.filesize / space + 1)

    def read(self, nbyte):
        byts = self.f.read(nbyte)
        if not (len(byts) == nbyte):
            raise EOFError('Reached the end of the file')
        return byts

    def findnext(self, do_cs=True):
        """Find the next data block by checking the checksum and the
        sync byte(0xa5)
        """
        sum = np.uint16(int('0xb58c', 0))  # Initialize the sum
        cs = 0
        func = _bitshift8
        func2 = np.uint8
        if self.endian == '<':
            func = np.uint8
            func2 = _bitshift8
        while True:
            val = unpack(self.endian + 'H', self.read(2))[0]
            if func(val) == 165 and (not do_cs or cs == np.uint16(sum)):
                self.f.seek(-2, 1)
                return hex(func2(val))
            sum += cs
            cs = val

    def read_id(self,):
        """Read the next 'ID' from the file.
        """
        self._thisid_bytes = bts = self.read(2)
        tmp = unpack(self.endian + 'BB', bts)
        if self.debug:
            logging.info('Position: {}, codes: {}'.format(self.f.tell(), tmp))
        if tmp[0] != 165:  # This catches a corrupted data block.
            if self.debug:
                logging.warning("Corrupted data block sync code (%d, %d) found "
                                "in ping %d. Searching for next valid code..." %
                                (tmp[0], tmp[1], self.c))
            val = int(self.findnext(do_cs=False), 0)
            self.f.seek(2, 1)
            if self.debug:
                logging.debug(
                    ' ...FOUND {} at position: {}.'.format(val, self.pos))
            return val
        return tmp[1]

    def readnext(self,):
        id = '0x%02x' % self.read_id()
        if id in self.fun_map:
            func_name = self.fun_map[id]
            out = getattr(self, func_name)()  # Should return None
            self._lastread = [func_name[5:]] + self._lastread[:-1]
            return out
        else:
            logging.warning('Unrecognized identifier: ' + id)
            self.f.seek(-2, 1)
            return 10

    def readfile(self, nlines=None):
        print('Reading file %s ...' % self.fname)
        retval = None
        try:
            while not retval:
                if self.c == nlines:
                    break
                retval = self.readnext()
                if retval == 10:
                    self.findnext()
                    retval = None
                if self._npings is not None and self.c >= self._npings:
                    if 'microstrain' in self._dtypes:
                        try:
                            self.readnext()
                        except:
                            pass
                    break
        except EOFError:
            if self.debug:
                logging.info(' end of file at {} bytes.'.format(self.pos))
        else:
            if self.debug:
                logging.info(' stopped at {} bytes.'.format(self.pos))
        self.c -= 1
        _crop_data(self.data, slice(0, self.c), self.n_samp_guess)

    def findnextid(self, id):
        if id.__class__ is str:
            id = int(id, 0)
        nowid = None
        while nowid != id:
            nowid = self.read_id()
            if nowid == 16:
                shift = 22
            else:
                sz = 2 * unpack(self.endian + 'H', self.read(2))[0]
                shift = sz - 4
            self.f.seek(shift, 1)
        return self.pos

    def code_spacing(self, searchcode, iternum=50):
        """
        Find the spacing, in bytes, between a specific hardware code.
        Repeat this * iternum * times(default 50).
        Returns the average spacing, in bytes, between the code.
        """
        p0 = self.findnextid(searchcode)
        for i in range(iternum):
            try:
                self.findnextid(searchcode)
            except EOFError:
                break
        if self.debug:
            logging.info('p0={}, pos={}, i={}'.format(p0, self.pos, i))
        # Compute the average of the data size:
        return (self.pos - p0) / (i + 1)

    def checksum(self, byts):
        """Perform a checksum on `byts` and read the checksum value.
        """
        if self.do_checksum:
            if not np.sum(unpack(self.endian + str(int(1 + len(byts) / 2)) + 'H',
                                 self._thisid_bytes + byts)) + \
                    46476 - unpack(self.endian + 'H', self.read(2)):

                raise Exception("CheckSum Failed at {}".format(self.pos))
        else:
            self.f.seek(2, 1)

    def read_user_cfg(self,):
        # ID: '0x00 = 00
        if self.debug:
            logging.info('Reading user configuration (0x00) ping #{} @ {}...'
                         .format(self.c, self.pos))
        cfg_u = self.config
        byts = self.read(508)
        # the first two bytes are the size.
        tmp = unpack(self.endian +
                     '2x18H6s4HI9H90H80s48xH50x6H4xH2x2H2xH30x8H',
                     byts)
        cfg_u['usr'] = {}
        cfg_u['adv'] = {}
        cfg_u['awac'] = {}

        cfg_u['transmit_pulse_length_m'] = tmp[0]  # counts
        cfg_u['blank_dist'] = tmp[1]  # overridden below
        cfg_u['receive_length_m'] = tmp[2]  # counts
        cfg_u['time_between_pings'] = tmp[3]  # counts
        cfg_u['time_between_bursts'] = tmp[4]  # counts
        cfg_u['adv']['n_pings_per_burst'] = tmp[5]
        cfg_u['awac']['avg_interval'] = tmp[6]
        cfg_u['usr']['n_beams'] = tmp[7]
        TimCtrlReg = _int2binarray(tmp[8], 16).astype(int)
        # From the nortek system integrator manual
        # (note: bit numbering is zero-based)
        cfg_u['usr']['profile_mode'] = [
            'single', 'continuous'][TimCtrlReg[1]]
        cfg_u['usr']['burst_mode'] = str(bool(~TimCtrlReg[2]))
        cfg_u['usr']['power_level'] = TimCtrlReg[5] + 2 * TimCtrlReg[6] + 1
        cfg_u['usr']['sync_out_pos'] = ['middle', 'end', ][TimCtrlReg[7]]
        cfg_u['usr']['sample_on_sync'] = str(bool(TimCtrlReg[8]))
        cfg_u['usr']['start_on_sync'] = str(bool(TimCtrlReg[9]))
        cfg_u['PwrCtrlReg'] = _int2binarray(tmp[9], 16)
        cfg_u['A1'] = tmp[10]
        cfg_u['B0'] = tmp[11]
        cfg_u['B1'] = tmp[12]
        cfg_u['usr']['compass_update_rate'] = tmp[13]
        cfg_u['coord_sys_axes'] = ['ENU', 'XYZ', 'beam'][tmp[14]]
        cfg_u['usr']['n_bins'] = tmp[15]
        cfg_u['bin_length'] = tmp[16]
        cfg_u['burst_interval'] = tmp[17]
        cfg_u['usr']['deployment_name'] = tmp[18].partition(b'\x00')[
            0].decode('utf-8')
        cfg_u['usr']['wrap_mode'] = str(bool(tmp[19]))
        cfg_u['deployment_time'] = np.array(tmp[20:23])
        cfg_u['diagnotics_interval'] = tmp[23]
        Mode0 = _int2binarray(tmp[24], 16)
        cfg_u['user_soundspeed_adj_factor'] = tmp[25]
        cfg_u['n_samples_diag'] = tmp[26]
        cfg_u['n_beams_cells_diag'] = tmp[27]
        cfg_u['n_pings_diag_wave'] = tmp[28]
        ModeTest = _int2binarray(tmp[29], 16)
        cfg_u['usr']['analog_in'] = tmp[30]
        sfw_ver = str(tmp[31])
        cfg_u['usr']['software_version'] = sfw_ver[0] + \
            '.'+sfw_ver[1:3]+'.'+sfw_ver[3:]
        cfg_u['usr']['salinity'] = tmp[32]/10
        cfg_u['VelAdjTable'] = np.array(tmp[33:123])
        cfg_u['usr']['comments'] = tmp[123].partition(b'\x00')[
            0].decode('utf-8')
        cfg_u['awac']['wave_processing_method'] = [
            'PUV', 'SUV', 'MLM', 'MLMST', 'None'][tmp[124]]
        Mode1 = _int2binarray(tmp[125], 16)
        cfg_u['awac']['prc_dyn_wave_cell_pos'] = int(tmp[126]/32767 * 100)
        cfg_u['wave_transmit_pulse'] = tmp[127]
        cfg_u['wave_blank_dist'] = tmp[128]
        cfg_u['awac']['wave_cell_size'] = tmp[129]
        cfg_u['awac']['n_samples_wave'] = tmp[130]
        cfg_u['n_burst'] = tmp[131]
        cfg_u['analog_out_scale'] = tmp[132]
        cfg_u['corr_thresh'] = tmp[133]
        cfg_u['transmit_pulse_lag2'] = tmp[134]  # counts
        cfg_u['QualConst'] = np.array(tmp[135:143])
        self.checksum(byts)
        cfg_u['usr']['user_specified_sound_speed'] = str(Mode0[0])
        cfg_u['awac']['wave_mode'] = ['Disabled', 'Enabled'][int(Mode0[1])]
        cfg_u['usr']['analog_output'] = str(Mode0[2])
        cfg_u['usr']['output_format'] = ['Vector', 'ADV'][int(Mode0[3])]  # noqa
        cfg_u['vel_scale_mm'] = [1, 0.1][int(Mode0[4])]
        cfg_u['usr']['serial_output'] = str(Mode0[5])
        cfg_u['reserved_EasyQ'] = str(Mode0[6])
        cfg_u['usr']['power_output_analog'] = str(Mode0[8])
        cfg_u['mode_test_use_DSP'] = str(ModeTest[0])
        cfg_u['mode_test_filter_output'] = ['total', 'correction_only'][int(ModeTest[1])]  # noqa
        cfg_u['awac']['wave_fs'] = ['1 Hz', '2 Hz'][int(Mode1[0])]
        cfg_u['awac']['wave_cell_position'] = ['fixed', 'dynamic'][int(Mode1[1])]  # noqa
        cfg_u['awac']['type_wave_cell_pos'] = ['pct_of_mean_pressure', 'pct_of_min_re'][int(Mode1[2])]  # noqa

    def read_head_cfg(self,):
        # ID: '0x04 = 04
        if self.debug:
            logging.info('Reading head configuration (0x04) ping #{} @ {}...'
                         .format(self.c, self.pos))
        cfg = self.config
        cfg['head'] = {}
        byts = self.read(220)
        tmp = unpack(self.endian + '2x3H12s176s22sH', byts)
        head_config = _int2binarray(tmp[0], 16).astype(int)
        cfg['head']['pressure_sensor'] = ['no', 'yes'][head_config[0]]
        cfg['head']['compass'] = ['no', 'yes'][head_config[1]]
        cfg['head']['tilt_sensor'] = ['no', 'yes'][head_config[2]]
        cfg['head']['carrier_freq_kHz'] = tmp[1]
        cfg['beam2inst_orientmat'] = np.array(
            unpack(self.endian + '9h', tmp[4][8:26])).reshape(3, 3) / 4096.
        self.checksum(byts)

    def read_hw_cfg(self,):
        # ID 0x05 = 05
        if self.debug:
            logging.info('Reading hardware configuration (0x05) ping #{} @ {}...'
                         .format(self.c, self.pos))
        cfg_hw = self.config
        cfg_hw['hdw'] = {}
        byts = self.read(44)
        tmp = unpack(self.endian + '2x14s6H12x4s', byts)
        cfg_hw['hdw']['serial_number'] = tmp[0][:8].decode('utf-8')
        cfg_hw['ProLogID'] = unpack('B', tmp[0][8:9])[0]
        cfg_hw['hdw']['ProLogFWver'] = tmp[0][10:].decode('utf-8')
        cfg_hw['board_config'] = tmp[1]
        cfg_hw['board_freq'] = tmp[2]
        cfg_hw['hdw']['PIC_version'] = tmp[3]
        cfg_hw['hdw']['hardware_rev'] = tmp[4]
        cfg_hw['hdw']['recorder_size_bytes'] = tmp[5] * 65536
        status = _int2binarray(tmp[6], 16).astype(int)
        cfg_hw['hdw']['vel_range'] = ['normal', 'high'][status[0]]
        cfg_hw['hdw']['firmware_version'] = tmp[7].decode('utf-8')
        self.checksum(byts)

    def rd_time(self, strng):
        """Read the time from the first 6bytes of the input string.
        """
        min, sec, day, hour, year, month = unpack('BBBBBB', strng[:6])
        return time.date2epoch(datetime(time._fullyear(_bcd2char(year)),
                                        _bcd2char(month),
                                        _bcd2char(day),
                                        _bcd2char(hour),
                                        _bcd2char(min),
                                        _bcd2char(sec)))[0]

    def _init_data(self, vardict):
        """Initialize the data object according to vardict.

        Parameters
        ----------
        vardict : (dict of :class:`<VarAttrs>`)
          The variable definitions in the :class:`<VarAttrs>` specify
          how to initialize each data variable.

        """
        shape_args = {'n': self.n_samp_guess}
        try:
            shape_args['nbins'] = self.config['usr']['n_bins']
        except KeyError:
            pass
        for nm, va in list(vardict.items()):
            if va.group is None:
                # These have to stay separated.
                if nm not in self.data:
                    self.data[nm] = va._empty_array(**shape_args)
            else:
                if nm not in self.data[va.group]:
                    self.data[va.group][nm] = va._empty_array(**shape_args)
                    self.data['units'][nm] = va.units

    def read_vec_data(self,):
        # ID: 0x10 = 16
        c = self.c
        dat = self.data
        if self.debug:
            logging.info('Reading vector velocity data (0x10) ping #{} @ {}...'
                         .format(self.c, self.pos))

        if 'vel' not in dat['data_vars']:
            self._init_data(nortek_defs.vec_data)
            self._dtypes += ['vec_data']

        byts = self.read(20)
        ds = dat['sys']
        dv = dat['data_vars']
        (ds['AnaIn2LSB'][c],
         ds['Count'][c],
         dv['PressureMSB'][c],
         ds['AnaIn2MSB'][c],
         dv['PressureLSW'][c],
         ds['AnaIn1'][c],
         dv['vel'][0, c],
         dv['vel'][1, c],
         dv['vel'][2, c],
         dv['amp'][0, c],
         dv['amp'][1, c],
         dv['amp'][2, c],
         dv['corr'][0, c],
         dv['corr'][1, c],
         dv['corr'][2, c]) = unpack(self.endian + '4B2H3h6B', byts)

        self.checksum(byts)
        self.c += 1

    def read_vec_checkdata(self,):
        # ID: 0x07 = 07
        if self.debug:
            logging.info('Reading vector check data (0x07) ping #{} @ {}...'
                         .format(self.c, self.pos))
        byts0 = self.read(6)
        checknow = {}
        tmp = unpack(self.endian + '2x2H', byts0)  # The first two are size.
        checknow['Samples'] = tmp[0]
        n = checknow['Samples']
        checknow['First_samp'] = tmp[1]
        checknow['Amp1'] = tbx._nans(n, dtype=np.uint8) + 8
        checknow['Amp2'] = tbx._nans(n, dtype=np.uint8) + 8
        checknow['Amp3'] = tbx._nans(n, dtype=np.uint8) + 8
        byts1 = self.read(3 * n)
        tmp = unpack(self.endian + (3 * n * 'B'), byts1)
        for idx, nm in enumerate(['Amp1', 'Amp2', 'Amp3']):
            checknow[nm] = np.array(tmp[idx * n:(idx + 1) * n], dtype=np.uint8)
        self.checksum(byts0 + byts1)
        if 'checkdata' not in self.config:
            self.config['checkdata'] = checknow
        else:
            if not isinstance(self.config['checkdata'], list):
                self.config['checkdata'] = [self.config['checkdata']]
            self.config['checkdata'] += [checknow]

    def _sci_data(self, vardict):
        """Convert the data to scientific units accordint to vardict.

        Parameters
        ----------
        vardict : (dict of :class:`<VarAttrs>`)
          The variable definitions in the :class:`<VarAttrs>` specify
          how to scale each data variable.

        """
        for nm, vd in list(vardict.items()):
            if vd.group is None:
                dat = self.data
            else:
                dat = self.data[vd.group]
            retval = vd.sci_func(dat[nm])
            # This checks whether a new data object was created:
            # sci_func returns None if it modifies the existing data.
            if retval is not None:
                dat[nm] = retval

    def sci_vec_data(self,):
        self._sci_data(nortek_defs.vec_data)
        dat = self.data

        dat['data_vars']['pressure'] = (
            dat['data_vars']['PressureMSB'].astype('float32') * 65536 +
            dat['data_vars']['PressureLSW'].astype('float32')) / 1000.
        dat['units']['pressure'] = 'dbar'

        dat['data_vars'].pop('PressureMSB')
        dat['data_vars'].pop('PressureLSW')

        # Apply velocity scaling (1 or 0.1)
        dat['data_vars']['vel'] *= self.config['vel_scale_mm']

    def read_vec_hdr(self,):
        # ID: '0x12 = 18
        if self.debug:
            logging.info('Reading vector header data (0x12) ping #{} @ {}...'
                         .format(self.c, self.pos))
        byts = self.read(38)
        # The first two are size, the next 6 are time.
        tmp = unpack(self.endian + '8xH7B21x', byts)
        hdrnow = {}
        hdrnow['time'] = self.rd_time(byts[2:8])
        hdrnow['NRecords'] = tmp[0]
        hdrnow['Noise1'] = tmp[1]
        hdrnow['Noise2'] = tmp[2]
        hdrnow['Noise3'] = tmp[3]
        hdrnow['Spare0'] = byts[13:14].decode('utf-8')
        hdrnow['Corr1'] = tmp[5]
        hdrnow['Corr2'] = tmp[6]
        hdrnow['Corr3'] = tmp[7]
        hdrnow['Spare1'] = byts[17:].decode('utf-8')
        self.checksum(byts)
        if 'data_header' not in self.config:
            self.config['data_header'] = hdrnow
        else:
            if not isinstance(self.config['data_header'], list):
                self.config['data_header'] = [self.config['data_header']]
            self.config['data_header'] += [hdrnow]

    def read_vec_sysdata(self,):
        # ID: 0x11 = 17
        c = self.c
        if self.debug:
            logging.info('Reading vector system data (0x11) ping #{} @ {}...'
                         .format(self.c, self.pos))
        dat = self.data
        if self._lastread[:2] == ['vec_checkdata', 'vec_hdr', ]:
            self.burst_start[c] = True
        if 'time' not in dat['coords']:
            self._init_data(nortek_defs.vec_sysdata)
            self._dtypes += ['vec_sysdata']
        byts = self.read(24)
        # The first two are size (skip them).
        dat['coords']['time'][c] = self.rd_time(byts[2:8])
        ds = dat['sys']
        dv = dat['data_vars']
        (dv['batt'][c],
         dv['c_sound'][c],
         dv['heading'][c],
         dv['pitch'][c],
         dv['roll'][c],
         dv['temp'][c],
         dv['error'][c],
         dv['status'][c],
         ds['AnaIn'][c]) = unpack(self.endian + '2H3hH2BH', byts[8:])
        self.checksum(byts)

    def sci_vec_sysdata(self,):
        """Translate the data in the vec_sysdata structure into
        scientific units.
        """
        dat = self.data
        fs = dat['attrs']['fs']
        self._sci_data(nortek_defs.vec_sysdata)
        t = dat['coords']['time']
        dv = dat['data_vars']
        dat['sys']['_sysi'] = ~np.isnan(t)
        # These are the indices in the sysdata variables
        # that are not interpolated.
        nburst = self.config['n_burst']
        dv['orientation_down'] = tbx._nans(len(t), dtype='bool')
        if nburst == 0:
            num_bursts = 1
            nburst = len(t)
        else:
            num_bursts = int(len(t) // nburst + 1)
        for nb in range(num_bursts):
            iburst = slice(nb * nburst, (nb + 1) * nburst)
            sysi = dat['sys']['_sysi'][iburst]
            if len(sysi) == 0:
                break
            # Skip the first entry for the interpolation process
            inds = np.nonzero(sysi)[0][1:]
            arng = np.arange(len(t[iburst]), dtype=np.float64)
            if len(inds) >= 2:
                p = np.poly1d(np.polyfit(inds, t[iburst][inds], 1))
                t[iburst] = p(arng)
            elif len(inds) == 1:
                t[iburst] = ((arng - inds[0]) / (fs * 3600 * 24) +
                             t[iburst][inds[0]])
            else:
                t[iburst] = (t[iburst][0] + arng / (fs * 24 * 3600))

            tmpd = tbx._nans_like(dv['heading'][iburst])
            # The first status bit should be the orientation.
            tmpd[sysi] = dv['status'][iburst][sysi] & 1
            tbx.fillgaps(tmpd, extrapFlg=True)
            tmpd = np.nan_to_num(tmpd, nan=0)  # nans in pitch roll heading
            slope = np.diff(tmpd)
            tmpd[1:][slope < 0] = 1
            tmpd[:-1][slope > 0] = 0
            dv['orientation_down'][iburst] = tmpd.astype('bool')
        tbx.interpgaps(dv['batt'], t)
        tbx.interpgaps(dv['c_sound'], t)
        tbx.interpgaps(dv['heading'], t)
        tbx.interpgaps(dv['pitch'], t)
        tbx.interpgaps(dv['roll'], t)
        tbx.interpgaps(dv['temp'], t)

    def read_microstrain(self,):
        """Read ADV microstrain sensor (IMU) data
        """
        # 0x71 = 113
        if self.c == 0:
            logging.warning('First "microstrain data" block '
                            'is before first "vector system data" block.')
        else:
            self.c -= 1
        if self.debug:
            logging.info('Reading vector microstrain data (0x71) ping #{} @ {}...'
                         .format(self.c, self.pos))
        byts0 = self.read(4)
        # The first 2 are the size, 3rd is count, 4th is the id.
        ahrsid = unpack(self.endian + '3xB', byts0)[0]
        if hasattr(self, '_ahrsid') and self._ahrsid != ahrsid:
            logging.warning('AHRS_ID changes mid-file!')

        if ahrsid in [195, 204, 210, 211]:
            self._ahrsid = ahrsid

        c = self.c
        dat = self.data
        dv = dat['data_vars']
        da = dat['attrs']
        da['has_imu'] = 1  # logical
        if 'accel' not in dv:
            self._dtypes += ['microstrain']
            if ahrsid == 195:
                self._orient_dnames = ['accel', 'angrt', 'orientmat']
                dv['accel'] = tbx._nans((3, self.n_samp_guess),
                                        dtype=np.float32)
                dv['angrt'] = tbx._nans((3, self.n_samp_guess),
                                        dtype=np.float32)
                dv['orientmat'] = tbx._nans((3, 3, self.n_samp_guess),
                                            dtype=np.float32)
                rv = ['accel', 'angrt']
                if not all(x in da['rotate_vars'] for x in rv):
                    da['rotate_vars'].extend(rv)
                dat['units'].update({'accel': 'm/s^2',
                                     'angrt': 'rad/s'})

            if ahrsid in [204, 210]:
                self._orient_dnames = ['accel', 'angrt', 'mag', 'orientmat']
                dv['accel'] = tbx._nans((3, self.n_samp_guess),
                                        dtype=np.float32)
                dv['angrt'] = tbx._nans((3, self.n_samp_guess),
                                        dtype=np.float32)
                dv['mag'] = tbx._nans((3, self.n_samp_guess),
                                      dtype=np.float32)
                rv = ['accel', 'angrt', 'mag']
                if not all(x in da['rotate_vars'] for x in rv):
                    da['rotate_vars'].extend(rv)
                if ahrsid == 204:
                    dv['orientmat'] = tbx._nans((3, 3, self.n_samp_guess),
                                                dtype=np.float32)
                dat['units'].update({'accel': 'm/s^2',
                                     'angrt': 'rad/s',
                                     'mag': 'gauss'})

            elif ahrsid == 211:
                self._orient_dnames = ['angrt', 'accel', 'mag']
                dv['angrt'] = tbx._nans((3, self.n_samp_guess),
                                        dtype=np.float32)
                dv['accel'] = tbx._nans((3, self.n_samp_guess),
                                        dtype=np.float32)
                dv['mag'] = tbx._nans((3, self.n_samp_guess),
                                      dtype=np.float32)
                rv = ['angrt', 'accel', 'mag']
                if not all(x in da['rotate_vars'] for x in rv):
                    da['rotate_vars'].extend(rv)
                dat['units'].update({'accel': 'm/s^2',
                                     'angrt': 'rad/s',
                                     'mag': 'gauss'})
        byts = ''
        if ahrsid == 195:  # 0xc3
            byts = self.read(64)
            dt = unpack(self.endian + '6f9f4x', byts)
            (dv['angrt'][:, c],
             dv['accel'][:, c]) = (dt[0:3], dt[3:6],)
            dv['orientmat'][:, :, c] = ((dt[6:9], dt[9:12], dt[12:15]))
        elif ahrsid == 204:  # 0xcc
            byts = self.read(78)
            # This skips the "DWORD" (4 bytes) and the AHRS checksum
            # (2 bytes)
            dt = unpack(self.endian + '18f6x', byts)
            (dv['accel'][:, c],
             dv['angrt'][:, c],
             dv['mag'][:, c]) = (dt[0:3], dt[3:6], dt[6:9],)
            dv['orientmat'][:, :, c] = ((dt[9:12], dt[12:15], dt[15:18]))
        elif ahrsid == 211:
            byts = self.read(42)
            dt = unpack(self.endian + '9f6x', byts)
            (dv['angrt'][:, c],
             dv['accel'][:, c],
             dv['mag'][:, c]) = (dt[0:3], dt[3:6], dt[6:9],)
        else:
            logging.warning('Unrecognized IMU identifier: ' + str(ahrsid))
            self.f.seek(-2, 1)
            return 10
        self.checksum(byts0 + byts)
        self.c += 1  # reset the increment

    def sci_microstrain(self,):
        """Rotate orientation data into ADV coordinate system.
        """
        # MS = MicroStrain
        dv = self.data['data_vars']
        for nm in self._orient_dnames:
            # Rotate the MS orientation data (in MS coordinate system)
            # to be consistent with the ADV coordinate system.
            # (x,y,-z)_ms = (z,y,x)_adv
            (dv[nm][2],
             dv[nm][0]) = (dv[nm][0],
                           -dv[nm][2].copy())
        if 'orientmat' in self._orient_dnames:
            # MS coordinate system is in North-East-Down (NED),
            # we want East-North-Up (ENU)
            dv['orientmat'][:, 2] *= -1
            (dv['orientmat'][:, 0],
             dv['orientmat'][:, 1]) = (dv['orientmat'][:, 1],
                                       dv['orientmat'][:, 0].copy())
        if 'accel' in dv:
            # This value comes from the MS 3DM-GX3 MIP manual
            dv['accel'] *= 9.80665
        if self._ahrsid in [195, 211]:
            # These are DAng and DVel, so we convert them to angrt, accel here
            dv['angrt'] *= self.config['fs']
            dv['accel'] *= self.config['fs']

    def read_awac_profile(self,):
        # ID: '0x20' = 32
        dat = self.data
        if self.debug:
            logging.info('Reading AWAC velocity data (0x20) ping #{} @ {}...'
                         .format(self.c, self.pos))
        nbins = self.config['usr']['n_bins']
        if 'temp' not in dat['data_vars']:
            self._init_data(nortek_defs.awac_profile)
            self._dtypes += ['awac_profile']

        # Note: docs state there is 'fill' byte at the end, if nbins is odd,
        # but doesn't appear to be the case
        n = self.config['usr']['n_beams']
        byts = self.read(116 + n*3 * nbins)
        c = self.c
        dat['coords']['time'][c] = self.rd_time(byts[2:8])
        ds = dat['sys']
        dv = dat['data_vars']
        (dv['error'][c],
         ds['AnaIn1'][c],
         dv['batt'][c],
         dv['c_sound'][c],
         dv['heading'][c],
         dv['pitch'][c],
         dv['roll'][c],
         p_msb,
         dv['status'][c],
         p_lsw,
         dv['temp'][c],) = unpack(self.endian + '7HBB2H', byts[8:28])
        dv['pressure'][c] = (65536 * p_msb + p_lsw)
        # The nortek system integrator manual specifies an 88byte 'spare'
        # field, therefore we start at 116.
        tmp = unpack(self.endian + str(n * nbins) + 'h' +
                     str(n * nbins) + 'B', byts[116:116 + n*3 * nbins])
        for idx in range(n):
            dv['vel'][idx, :, c] = tmp[idx * nbins: (idx + 1) * nbins]
            dv['amp'][idx, :, c] = tmp[(idx + n) * nbins: (idx + n+1) * nbins]
        self.checksum(byts)
        self.c += 1

    def sci_awac_profile(self,):
        self._sci_data(nortek_defs.awac_profile)
        # Calculate the ranges.
        cs_coefs = {2000: 0.0239,
                    1000: 0.0478,
                    600: 0.0797,
                    400: 0.1195}
        h_ang = 25 * (np.pi / 180)  # Head angle is 25 degrees for all awacs.
        # Cell size
        cs = round(float(self.config['bin_length']) / 256. *
                   cs_coefs[self.config['head']['carrier_freq_kHz']] * np.cos(h_ang), ndigits=2)
        # Blanking distance
        bd = round(self.config['blank_dist'] *
                   0.0229 * np.cos(h_ang) - cs, ndigits=2)

        r = (np.float32(np.arange(self.config['usr']['n_bins']))+1)*cs + bd
        self.data['coords']['range'] = r
        self.data['attrs']['cell_size'] = cs
        self.data['attrs']['blank_dist'] = bd

    def dat2sci(self,):
        for nm in self._dtypes:
            getattr(self, 'sci_' + nm)()
        for nm in ['data_header', 'checkdata']:
            if nm in self.config and isinstance(self.config[nm], list):
                self.config[nm] = _recatenate(self.config[nm])

    def __exit__(self, type, value, trace):
        self.close()

    def __enter__(self):
        return self


def _crop_data(obj, range, n_lastdim):
    for nm, dat in obj.items():
        if isinstance(dat, np.ndarray) and (dat.shape[-1] == n_lastdim):
            obj[nm] = dat[..., range]


def _recatenate(obj):
    out = type(obj[0])()
    for ky in list(obj[0].keys()):
        if ky in ['__data_groups__', '_type']:
            continue
        val0 = obj[0][ky]
        if isinstance(val0, np.ndarray) and val0.size > 1:
            out[ky] = np.concatenate([val[ky][..., None] for val in obj],
                                     axis=-1)
        else:
            out[ky] = np.array([val[ky] for val in obj])
    return out



import numpy as np
import xarray as xr
from struct import unpack, calcsize
import warnings
from pathlib import Path
import logging

from . import nortek2_defs as defs
from . import nortek2_lib as lib
from .base import _find_userdata, _create_dataset, _abspath
from ..rotate.vector import _euler2orient
from ..rotate.base import _set_coords
from ..rotate.api import set_declination
from ..time import epoch2dt64, _fill_time_gaps


def read_signature(filename, userdata=True, nens=None, rebuild_index=False,
                   debug=False, **kwargs):
    """Read a Nortek Signature (.ad2cp) datafile

    Parameters
    ----------
    filename : string
        The filename of the file to load.
    userdata : bool
        To search for and use a .userdata.json or not
    nens : int, or tuple of 2 ints
        The number of ensembles to read, if int (starting at the
        beginning); or the range of ensembles to read, if tuple.

    Returns
    -------
    ds : xarray.Dataset
        An xarray dataset from the binary instrument data

    """
    # Start debugger logging
    if debug:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
        filepath = Path(filename)
        logfile = filepath.with_suffix('.log')
        logging.basicConfig(filename=str(logfile),
                            filemode='w',
                            level=logging.NOTSET,
                            format='%(name)s - %(levelname)s - %(message)s')

    if nens is None:
        nens = [0, None]
    else:
        try:
            n = len(nens)
        except TypeError:
            nens = [0, nens]
        else:
            # passes: it's a list/tuple/array
            if n != 2:
                raise TypeError('nens must be: None (), int, or len 2')

    userdata = _find_userdata(filename, userdata)

    rdr = _Ad2cpReader(filename, rebuild_index=rebuild_index, debug=debug)
    d = rdr.readfile(nens[0], nens[1])
    rdr.sci_data(d)
    out = _reorg(d)
    _reduce(out)

    # Convert time to dt64 and fill gaps
    coords = out['coords']
    t_list = [t for t in coords if 'time' in t]
    for ky in t_list:
        tdat = coords[ky]
        tdat[tdat == 0] = np.NaN
        if np.isnan(tdat).any():
            tag = ky.lstrip('time')
            warnings.warn("Zero/NaN values found in '{}'. Interpolating and "
                          "extrapolating them. To identify which values were filled later, "
                          "look for 0 values in 'status{}'".format(ky, tag))
            tdat = _fill_time_gaps(tdat, sample_rate_hz=out['attrs']['fs'])
        coords[ky] = epoch2dt64(tdat).astype('datetime64[us]')

    declin = None
    for nm in userdata:
        if 'dec' in nm:
            declin = userdata[nm]
        else:
            out['attrs'][nm] = userdata[nm]

    # Create xarray dataset from upper level dictionary
    ds = _create_dataset(out)
    ds = _set_coords(ds, ref_frame=ds.coord_sys)

    if 'orientmat' not in ds:
        omat = _euler2orient(ds['heading'], ds['pitch'], ds['roll'])
        ds['orientmat'] = xr.DataArray(omat,
                                       coords={'earth': ['E', 'N', 'U'],
                                               'inst': ['X', 'Y', 'Z'],
                                               'time': ds['time']},
                                       dims=['earth', 'inst', 'time'])
    if declin is not None:
        set_declination(ds, declin, inplace=True)

    # Close handler
    if debug:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
            handler.close()

    return ds


class _Ad2cpReader():
    def __init__(self, fname, endian=None, bufsize=None, rebuild_index=False,
                 debug=False):
        self.fname = fname
        self._check_nortek(endian)
        self.f.seek(0, 2)  # Seek to end
        self._eof = self.f.tell()
        self._index = lib.get_index(fname,
                                    reload=rebuild_index,
                                    debug=debug)
        self._reopen(bufsize)
        self.filehead_config = self._read_filehead_config_string()
        self._ens_pos = self._index['pos'][lib._boolarray_firstensemble_ping(
            self._index)]
        self._lastblock_iswhole = self._calc_lastblock_iswhole()
        self._config = lib._calc_config(self._index)
        self._init_burst_readers()
        self.unknown_ID_count = {}

    def _calc_lastblock_iswhole(self, ):
        blocksize, blocksize_count = np.unique(np.diff(self._ens_pos),
                                               return_counts=True)
        standard_blocksize = blocksize[blocksize_count.argmax()]
        return (self._eof - self._ens_pos[-1]) == standard_blocksize

    def _check_nortek(self, endian):
        self._reopen(10)
        byts = self.f.read(2)
        if endian is None:
            if unpack('<' + 'BB', byts) == (165, 10):
                endian = '<'
            elif unpack('>' + 'BB', byts) == (165, 10):
                endian = '>'
            else:
                raise Exception(
                    "I/O error: could not determine the 'endianness' "
                    "of the file.  Are you sure this is a Nortek "
                    "AD2CP file?")
        self.endian = endian

    def _reopen(self, bufsize=None):
        if bufsize is None:
            bufsize = 1000000
        try:
            self.f.close()
        except AttributeError:
            pass
        self.f = open(_abspath(self.fname), 'rb', bufsize)

    def _read_filehead_config_string(self, ):
        hdr = self._read_hdr()
        out = {}
        s_id, string = self._read_str(hdr['sz'])
        string = string.decode('utf-8')
        for ln in string.splitlines():
            ky, val = ln.split(',', 1)
            if ky in out:
                # There are more than one of this key
                if not isinstance(out[ky], list):
                    tmp = out[ky]
                    out[ky] = []
                    out[ky].append(tmp)
                out[ky].append(val)
            else:
                out[ky] = val
        out2 = {}
        for ky in out:
            if ky.startswith('GET'):
                dat = out[ky]
                d = out2[ky.lstrip('GET')] = dict()
                for itm in dat.split(','):
                    k, val = itm.split('=')
                    try:
                        val = int(val)
                    except ValueError:
                        try:
                            val = float(val)
                        except ValueError:
                            pass
                    d[k] = val
            else:
                out2[ky] = out[ky]
        return out2

    def _init_burst_readers(self, ):
        self._burst_readers = {}
        for rdr_id, cfg in self._config.items():
            if rdr_id == 28:
                self._burst_readers[rdr_id] = defs._calc_echo_struct(
                    cfg['_config'], cfg['n_cells'])
            elif rdr_id == 23:
                self._burst_readers[rdr_id] = defs._calc_bt_struct(
                    cfg['_config'], cfg['n_beams'])
            else:
                self._burst_readers[rdr_id] = defs._calc_burst_struct(
                    cfg['_config'], cfg['n_beams'], cfg['n_cells'])

    def init_data(self, ens_start, ens_stop):
        outdat = {}
        nens = int(ens_stop - ens_start)
        n26 = ((self._index['ID'] == 26) &
               (self._index['ens'] >= ens_start) &
               (self._index['ens'] < ens_stop)).sum()
        for ky in self._burst_readers:
            if ky == 26:
                n = n26
                ens = np.zeros(n, dtype='uint32')
            else:
                ens = np.arange(ens_start,
                                ens_stop).astype('uint32')
                n = nens
            outdat[ky] = self._burst_readers[ky].init_data(n)
            outdat[ky]['ensemble'] = ens
            outdat[ky]['units'] = self._burst_readers[ky].data_units()
        return outdat

    def _read_hdr(self, do_cs=False):
        res = defs.header.read2dict(self.f, cs=do_cs)
        if res['sync'] != 165:
            raise Exception("Out of sync!")
        return res

    def _read_str(self, size):
        string = self.f.read(size)
        id = string[0]
        string = string[1:-1]
        return id, string

    def _read_burst(self, id, dat, c, echo=False):
        rdr = self._burst_readers[id]
        rdr.read_into(self.f, dat, c)

    def readfile(self, ens_start=0, ens_stop=None):
        # If the lastblock is not whole, we don't read it.
        # If it is, we do (don't subtract 1)
        nens_total = len(self._ens_pos) - int(not self._lastblock_iswhole)
        if ens_stop is None or ens_stop > nens_total:
            ens_stop = nens_total
        ens_start = int(ens_start)
        ens_stop = int(ens_stop)
        nens = ens_stop - ens_start
        outdat = self.init_data(ens_start, ens_stop)
        outdat['filehead_config'] = self.filehead_config
        print('Reading file %s ...' % self.fname)
        c = 0
        c26 = 0
        self.f.seek(self._ens_pos[ens_start], 0)
        while True:
            try:
                hdr = self._read_hdr()
            except IOError:
                return outdat
            id = hdr['id']
            if id in [21, 23, 24, 28]:  # vel, bt, vel_b5, echo
                self._read_burst(id, outdat[id], c)
            elif id in [26]:  # alt_raw (altimeter burst)
                rdr = self._burst_readers[26]
                if not hasattr(rdr, '_nsamp_index'):
                    first_pass = True
                    tmp_idx = rdr._nsamp_index = rdr._names.index('altraw_nsamp')  # noqa
                    shift = rdr._nsamp_shift = calcsize(
                        defs._format(rdr._format[:tmp_idx],
                                     rdr._N[:tmp_idx]))
                else:
                    first_pass = False
                    tmp_idx = rdr._nsamp_index
                    shift = rdr._nsamp_shift
                tmp_idx = tmp_idx + 2  # Don't add in-place
                self.f.seek(shift, 1)
                # Now read the num_samples
                sz = unpack('<I', self.f.read(4))[0]
                self.f.seek(-shift - 4, 1)
                if first_pass:
                    # Fix the reader
                    rdr._shape[tmp_idx].append(sz)
                    rdr._N[tmp_idx] = sz
                    rdr._struct = defs.Struct('<' + rdr.format)
                    rdr.nbyte = calcsize(rdr.format)
                    rdr._cs_struct = defs.Struct(
                        '<' + '{}H'.format(int(rdr.nbyte // 2)))
                    # Initialize the array
                    outdat[26]['altraw_samp'] = defs._nans(
                        [rdr._N[tmp_idx],
                         len(outdat[26]['altraw_samp'])],
                        dtype=np.uint16)
                else:
                    if sz != rdr._N[tmp_idx]:
                        raise Exception(
                            "The number of samples in this 'Altimeter Raw' "
                            "burst is different from prior bursts.")
                self._read_burst(id, outdat[id], c26)
                outdat[id]['ensemble'][c26] = c
                c26 += 1

            elif id in [22, 27, 29, 30, 31, 35, 36]:  # avg record, bt record,
                # DVL, alt record, avg alt_raw record, raw echo, raw echo transmit
                if self.debug:
                    logging.debug(
                        "Skipped ID: 0x{:02X} ({:02d})\n".format(id, id))
                self.f.seek(hdr['sz'], 1)
            elif id == 160:
                # 0xa0 (i.e., 160) is a 'string data record'
                if id not in outdat:
                    outdat[id] = dict()
                s_id, s = self._read_str(hdr['sz'], )
                outdat[id][(c, s_id)] = s
            else:
                if id not in self.unknown_ID_count:
                    self.unknown_ID_count[id] = 1
                    if self.debug:
                        logging.warning('Unknown ID: 0x{:02X}!'.format(id))
                else:
                    self.unknown_ID_count[id] += 1
                self.f.seek(hdr['sz'], 1)

            c = self._advance_ens_count(c, ens_start, nens_total)

            if c >= nens:
                return outdat

    def _advance_ens_count(self, c, ens_start, nens_total):
        """This method advances the counter when appropriate to do so.
        """
        # It's unfortunate that all of this count checking is so
        # complex, but this is the best I could come up with right
        # now.
        try:
            # Checks to makes sure we're not already at the end of the
            # self._ens_pos array
            _posnow = self._ens_pos[c + ens_start + 1]
        except IndexError:
            # We are at the end of the array, set _posnow
            # We use "+1" here because we want the >= in the while
            # loop to fail for this case so that we go ahead and read
            # the next ping without advancing the ens counter.
            _posnow = self._eof + 1
        while (self.f.tell() >= _posnow):
            c += 1
            if c + ens_start + 1 >= nens_total:
                # Again check end of count list
                break
            try:
                # Same check as above.
                _posnow = self._ens_pos[c + ens_start + 1]
            except IndexError:
                _posnow = self._eof + 1
        return c

    def sci_data(self, dat):
        for id in dat:
            dnow = dat[id]
            if id not in self._burst_readers:
                continue
            rdr = self._burst_readers[id]
            rdr.sci_data(dnow)
            if 'vel' in dnow and 'vel_scale' in dnow:
                dnow['vel'] = (dnow['vel'] *
                               10.0 ** dnow['vel_scale']).astype('float32')

    def __exit__(self, type, value, trace,):
        self.f.close()

    def __enter__(self,):
        return self


def _reorg(dat):
    """This function grabs the data from the dictionary of data types
    (organized by ID), and combines them into a single dictionary.
    """
    outdat = {'data_vars': {}, 'coords': {}, 'attrs': {},
              'units': {}, 'sys': {}, 'altraw': {}}
    cfg = outdat['attrs']
    cfh = cfg['filehead_config'] = dat['filehead_config']
    cfg['inst_model'] = (cfh['ID'].split(',')[0][5:-1])
    cfg['inst_make'] = 'Nortek'
    cfg['inst_type'] = 'ADCP'
    cfg['rotate_vars'] = ['vel', ]

    for id, tag in [(21, ''), (23, '_bt'), (24, '_b5'), (26, '_ast'), (28, '_echo')]:
        if id in [24, 26]:
            collapse_exclude = [0]
        else:
            collapse_exclude = []
        if id not in dat:
            continue
        dnow = dat[id]
        outdat['units'].update(dnow['units'])
        cfg['burst_config' + tag] = lib._headconfig_int2dict(
            lib._collapse(dnow['config'], exclude=collapse_exclude,
                          name='config'))
        outdat['coords']['time' + tag] = lib._calc_time(
            dnow['year'] + 1900,
            dnow['month'],
            dnow['day'],
            dnow['hour'],
            dnow['minute'],
            dnow['second'],
            dnow['usec100'].astype('uint32') * 100)
        tmp = lib._beams_cy_int2dict(
            lib._collapse(dnow['beam_config'], exclude=collapse_exclude,
                          name='beam_config'), 21)
        cfg['n_cells' + tag] = tmp['n_cells']
        cfg['coord_sys_axes' + tag] = tmp['cy']
        cfg['n_beams' + tag] = tmp['n_beams']
        cfg['ambig_vel' +
            tag] = lib._collapse(dnow['ambig_vel'], name='ambig_vel')

        for ky in ['SerialNum', 'cell_size', 'blank_dist', 'nominal_corr',
                   'power_level_dB']:
            cfg[ky + tag] = lib._collapse(dnow[ky],
                                          exclude=collapse_exclude,
                                          name=ky)

        for ky in ['c_sound', 'temp', 'pressure', 'heading', 'pitch', 'roll',
                   'mag', 'accel', 'batt', 'temp_clock', 'error',
                   'status', 'ensemble',
                   ]:
            outdat['data_vars'][ky + tag] = dnow[ky]
            if 'ensemble' in ky:
                outdat['data_vars'][ky + tag] += 1
                outdat['units'][ky + tag] = '#'

        for ky in ['vel', 'amp', 'corr', 'prcnt_gd', 'echo', 'dist',
                   'orientmat', 'angrt', 'quaternions', 'ast_pressure',
                   'alt_dist', 'alt_quality', 'alt_status',
                   'ast_dist', 'ast_quality', 'ast_offset_time',
                   'altraw_nsamp', 'altraw_dsamp', 'altraw_samp',
                   'status0', 'fom', 'temp_press', 'press_std',
                   'pitch_std', 'roll_std', 'heading_std', 'xmit_energy',
                   ]:
            if ky in dnow:
                outdat['data_vars'][ky + tag] = dnow[ky]

    # Move 'altimeter raw' data to its own down-sampled structure
    if 26 in dat:
        ard = outdat['altraw']
        for ky in list(outdat['data_vars']):
            if ky.endswith('_ast'):
                grp = ky.split('.')[0]
                if '.' in ky and grp not in ard:
                    ard[grp] = {}
                ard[ky.rstrip('_ast')] = outdat['data_vars'].pop(ky)

        # Read altimeter status
        alt_status = lib._alt_status2data(outdat['data_vars']['alt_status'])
        for ky in alt_status:
            outdat['attrs'][ky] = lib._collapse(
                alt_status[ky].astype('uint8'), name=ky)
        outdat['data_vars'].pop('alt_status')

        # Power level index
        power = {0: 'high', 1: 'med-high', 2: 'med-low', 3: 'low'}
        outdat['attrs']['power_level_alt'] = power[outdat['attrs'].pop(
            'power_level_idx_alt')]

    # Read status data
    status0_vars = [x for x in outdat['data_vars'] if 'status0' in x]
    # Status data is the same across all tags, and there is always a 'status' and 'status0'
    status0_key = status0_vars[0]
    status0_data = lib._status02data(outdat['data_vars'][status0_key])
    status_key = status0_key.replace('0', '')
    status_data = lib._status2data(outdat['data_vars'][status_key])

    # Individual status codes
    # Wake up state
    wake = {0: 'bad power', 1: 'power on', 2: 'break', 3: 'clock'}
    outdat['attrs']['wakeup_state'] = wake[lib._collapse(
        status_data.pop('wakeup_state'), name=ky)]

    # Instrument direction
    # 0: XUP, 1: XDOWN, 2: YUP, 3: YDOWN, 4: ZUP, 5: ZDOWN,
    # 7: AHRS, handle as ZUP
    nortek_orient = {0: 'horizontal', 1: 'horizontal', 2: 'horizontal',
                     3: 'horizontal', 4: 'up', 5: 'down', 7: 'AHRS'}
    outdat['attrs']['orientation'] = nortek_orient[lib._collapse(
        status_data.pop('orient_up'), name='orientation')]

    # Orientation detection
    orient_status = {0: 'fixed', 1: 'auto_UD', 3: 'AHRS-3D'}
    outdat['attrs']['orient_status'] = orient_status[lib._collapse(
        status_data.pop('auto_orientation'), name='orient_status')]

    # Status variables
    for ky in ['low_volt_skip', 'active_config', 'telemetry_data', 'boost_running']:
        outdat['data_vars'][ky] = status_data[ky].astype('uint8')

    # Processor idle state - need to save as 1/0 per netcdf attribute limitations
    for ky in status0_data:
        outdat['attrs'][ky] = lib._collapse(
            status0_data[ky].astype('uint8'), name=ky)

    # Remove status0 variables - keep status variables as they useful for finding missing pings
    [outdat['data_vars'].pop(var) for var in status0_vars]

    # Set coordinate system
    outdat['attrs']['coord_sys'] = {'XYZ': 'inst',
                                    'ENU': 'earth',
                                    'beam': 'beam'}[cfg['coord_sys_axes']]

    # Copy appropriate vars to rotate_vars
    for ky in ['accel', 'angrt', 'mag']:
        for dky in outdat['data_vars'].keys():
            if dky == ky or dky.startswith(ky + '_'):
                outdat['attrs']['rotate_vars'].append(dky)
    if 'vel_bt' in outdat['data_vars']:
        outdat['attrs']['rotate_vars'].append('vel_bt')

    return outdat


def _reduce(data):
    """This function takes the output from `reorg`, and further simplifies the
    data. Mostly this is combining system, environmental, and orientation data
    --- from different data structures within the same ensemble --- by
    averaging.
    """
    dv = data['data_vars']
    dc = data['coords']
    da = data['attrs']

    # Average these fields
    for ky in ['c_sound', 'temp', 'pressure',
               'temp_press', 'temp_clock', 'batt']:
        lib._reduce_by_average(dv, ky, ky + '_b5')

    # Angle-averaging is treated separately
    for ky in ['heading', 'pitch', 'roll']:
        lib._reduce_by_average_angle(dv, ky, ky + '_b5')

    dc['range'] = ((np.arange(dv['vel'].shape[1])+1) *
                   da['cell_size'] +
                   da['blank_dist'])
    if 'vel_b5' in dv:
        dc['range_b5'] = ((np.arange(dv['vel_b5'].shape[1])+1) *
                          da['cell_size_b5'] +
                          da['blank_dist_b5'])
    if 'echo_echo' in dv:
        dv['echo'] = dv.pop('echo_echo')
        dc['range_echo'] = ((np.arange(dv['echo'].shape[0])+1) *
                            da['cell_size_echo'] +
                            da['blank_dist_echo'])

    if 'orientmat' in data['data_vars']:
        da['has_imu'] = 1  # logical
        # Signature AHRS rotation matrix returned in "inst->earth"
        # Change to dolfyn's "earth->inst"
        dv['orientmat'] = np.rollaxis(dv['orientmat'], 1)
    else:
        da['has_imu'] = 0

    da['fs'] = da['filehead_config']['BURST'].pop('SR')
    tmat = da['filehead_config'].pop('XFBURST')
    tm = np.zeros((tmat['ROWS'], tmat['COLS']), dtype=np.float32)
    for irow in range(tmat['ROWS']):
        for icol in range(tmat['COLS']):
            tm[irow, icol] = tmat['M' + str(irow + 1) + str(icol + 1)]
    dv['beam2inst_orientmat'] = tm



import numpy as np
import xarray as xr
import warnings
from os.path import getsize
from pathlib import Path
import logging

from .rdi_lib import bin_reader
from . import rdi_defs as defs
from .base import _find_userdata, _create_dataset, _abspath
from .. import time as tmlib
from ..rotate.rdi import _calc_beam_orientmat, _calc_orientmat
from ..rotate.base import _set_coords
from ..rotate.api import set_declination


def read_rdi(filename, userdata=None, nens=None, debug_level=-1,
             vmdas_search=False, winriver=False, **kwargs):
    """
    Read a TRDI binary data file.

    Parameters
    ----------
    filename : string
      Filename of TRDI file to read.
    userdata : True, False, or string of userdata.json filename (default ``True``)
      Whether to read the '<base-filename>.userdata.json' file.
    nens : int, (default: None, read entire file)
      Number of pings to read from the file
    debug_level : int (default=-1)
      Debug level [0 - 2]
    vmdas_search : bool (default=False)
      Search from the end of each ensemble for the VMDAS navigation
      block.  The byte offsets are sometimes incorrect.
    winriver : bool (default=False)
      If file is winriver or not. Automatically set by dolfyn, this is helpful 
      for debugging

    Returns
    -------
    ds : xarray.Dataset
      An xarray dataset from the binary instrument data

    """
    # Start debugger logging
    if debug_level >= 0:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
        filepath = Path(filename)
        logfile = filepath.with_suffix('.log')
        logging.basicConfig(filename=str(logfile),
                            filemode='w',
                            level=logging.NOTSET,
                            format='%(name)s - %(levelname)s - %(message)s')

    # Reads into a dictionary of dictionaries using netcdf naming conventions
    # Should be easier to debug
    with _RDIReader(filename,
                    debug_level=debug_level,
                    vmdas_search=vmdas_search,
                    winriver=winriver) as ldr:
        datNB, datBB = ldr.load_data(nens=nens)

    dats = [dat for dat in [datNB, datBB] if dat is not None]

    # Read in userdata
    userdata = _find_userdata(filename, userdata)
    dss = []
    for dat in dats:
        for nm in userdata:
            dat['attrs'][nm] = userdata[nm]

        if 'time_gps' in dat['coords']:
            # GPS data not necessarily sampling at the same rate as ADCP DAQ.
            dat = _remove_gps_duplicates(dat)

        # Create xarray dataset from upper level dictionary
        # print('Time in dat', dat['coords']['time'])
        if not np.isfinite(dat['coords']['time'][0]):
            continue
        ds = _create_dataset(dat)
        ds = _set_coords(ds, ref_frame=ds.coord_sys)

        if 'hdwtime_gps' in ds:
            ds.hdwtime_gps.encoding['units'] = 'seconds since 1970-01-01 00:00:00'

        # Create orientation matrices
        if 'beam2inst_orientmat' not in ds:
            ds['beam2inst_orientmat'] = xr.DataArray(_calc_beam_orientmat(
                ds.beam_angle,
                ds.beam_pattern == 'convex'),
                coords={'x': [1, 2, 3, 4],
                        'x*': [1, 2, 3, 4]},
                dims=['x', 'x*'])

        if 'orientmat' not in ds:
            ds['orientmat'] = xr.DataArray(_calc_orientmat(ds),
                                           coords={'earth': ['E', 'N', 'U'],
                                                   'inst': ['X', 'Y', 'Z'],
                                                   'time': ds['time']},
                                           dims=['earth', 'inst', 'time'])

        # Check magnetic declination if provided via software and/or userdata
        _set_rdi_declination(ds, filename, inplace=True)

        # VMDAS applies gps correction on velocity in .ENX files only
        if filename.rsplit('.')[-1] == 'ENX':
            ds.attrs['vel_gps_corrected'] = 1
        else:  # (not ENR or ENS) or WinRiver files
            ds.attrs['vel_gps_corrected'] = 0

        # Convert time coords to dt64
        t_coords = [t for t in ds.coords if 'time' in t]
        for ky in t_coords:
            dt = tmlib.epoch2dt64(ds[ky])
            ds = ds.assign_coords({ky: dt})

        # Convert time vars to dt64
        t_data = [t for t in ds.data_vars if 'time' in t]
        for ky in t_data:
            dt = tmlib.epoch2dt64(ds[ky])
            ds[ky].data = dt
        dss += [ds]

    if len(dss) == 2:
        warnings.warn("\nTwo profiling configurations retrieved from file"
                      "\nReturning first.")

    # Close handler
    if debug_level >= 0:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
            handler.close()

    return dss[0]


def _remove_gps_duplicates(dat):
    """
    Removes duplicate and nan timestamp values in 'time_gps' coordinate,
    and add hardware (ADCP DAQ) timestamp corresponding to GPS acquisition
    (in addition to the GPS unit's timestamp).
    """

    dat['data_vars']['hdwtime_gps'] = dat['coords']['time']

    # Remove duplicate timestamp values, if applicable
    dat['coords']['time_gps'], idx = np.unique(dat['coords']['time_gps'],
                                               return_index=True)
    # Remove nan values, if applicable
    nan = np.zeros(dat['coords']['time'].shape, dtype=bool)
    if any(np.isnan(dat['coords']['time_gps'])):
        nan = np.isnan(dat['coords']['time_gps'])
        dat['coords']['time_gps'] = dat['coords']['time_gps'][~nan]

    for key in dat['data_vars']:
        if ('gps' in key) or ('nmea' in key):
            dat['data_vars'][key] = dat['data_vars'][key][idx]
            if sum(nan) > 0:
                dat['data_vars'][key] = dat['data_vars'][key][~nan]

    return dat


def _set_rdi_declination(dat, fname, inplace):
    """
    If magnetic_var_deg is set, this means that the declination is already
    included in the heading and in the velocity data.
    """

    declin = dat.attrs.pop('declination', None)  # userdata declination

    if dat.attrs['magnetic_var_deg'] != 0:  # from TRDI software if set
        dat.attrs['declination'] = dat.attrs['magnetic_var_deg']
        dat.attrs['declination_in_orientmat'] = 1  # logical

    if dat.attrs['magnetic_var_deg'] != 0 and declin is not None:
        warnings.warn(
            "'magnetic_var_deg' is set to {:.2f} degrees in the binary "
            "file '{}', AND 'declination' is set in the 'userdata.json' "
            "file. DOLfYN WILL USE THE VALUE of {:.2f} degrees in "
            "userdata.json. If you want to use the value in "
            "'magnetic_var_deg', delete the value from userdata.json and "
            "re-read the file."
            .format(dat.attrs['magnetic_var_deg'], fname, declin))
        dat.attrs['declination'] = declin

    if declin is not None:
        set_declination(dat, declin, inplace)


class _RDIReader():
    _pos = 0
    progress = 0
    _cfac = 180 / 2 ** 31
    _source = 0
    _fixoffset = 0
    _nbyte = 0
    _search_num = 30000  # Maximum distance? to search
    _debug7f79 = None

    def __init__(self, fname, navg=1, debug_level=0, vmdas_search=False, winriver=False):
        self.fname = _abspath(fname)
        print('\nReading file {} ...'.format(fname))
        self._debug_level = debug_level
        self._vmdas_search = vmdas_search
        self._winrivprob = winriver
        self.flag = 0
        self.cfg = {}
        self.cfgbb = {}
        self.hdr = {}
        self.f = bin_reader(self.fname)

        # Check header, double buffer, and get filesize
        self._filesize = getsize(self.fname)
        space = self.code_spacing()  # '0x7F'
        self._npings = int(self._filesize / (space + 2))
        if self._debug_level >= 0:
            logging.info('Done: {}'.format(self.cfg))
            logging.info('self._bb {}'.format(self._bb))
            logging.info(self.cfgbb)
        self.f.seek(self._pos, 0)
        self.n_avg = navg

        self.ensemble = defs._ensemble(self.n_avg, self.cfg['n_cells'])
        if self._bb:
            self.ensembleBB = defs._ensemble(self.n_avg, self.cfgbb['n_cells'])

        self.vars_read = defs._variable_setlist(['time'])
        if self._bb:
            self.vars_readBB = defs._variable_setlist(['time'])

        if self._debug_level >= 0:
            logging.info('  %d pings estimated in this file' % self._npings)

    def code_spacing(self, iternum=50):
        """
        Returns the average spacing, in bytes, between pings.
        Repeat this * iternum * times(default 50).
        """
        fd = self.f
        p0 = self._pos
        # Get basic header data and check dual profile
        if not self.read_hdr():
            raise RuntimeError('No header in this file')
        self._bb = self.check_for_double_buffer()

        # Turn off debugging to check code spacing
        debug_level = self._debug_level
        self._debug_level = -1
        for i in range(iternum):
            try:
                self.read_hdr()
            except:
                break
        # Compute the average of the data size:
        size = (self._pos - p0) / (i+1) * 0.995
        self.f = fd
        self._pos = p0
        self._debug_level = debug_level
        return size

    def read_hdr(self,):
        fd = self.f
        cfgid = list(fd.read_ui8(2))
        nread = 0
        if self._debug_level >= 0:
            logging.info('pos {}'.format(self.f.pos))
            logging.info('cfgid0: [{:x}, {:x}]'.format(*cfgid))
        while (cfgid[0] != 127 or cfgid[1] != 127) or not self.checkheader():
            nextbyte = fd.read_ui8(1)
            if nextbyte is None:
                return False
            pos = fd.tell()
            nread += 1
            cfgid[1] = cfgid[0]
            cfgid[0] = nextbyte
            if not pos % 1000:
                if self._debug_level >= 0:
                    logging.info('  Still looking for valid cfgid at file '
                                 'position %d ...' % pos)
        self._pos = self.f.tell() - 2
        self.read_hdrseg()
        return True

    def check_for_double_buffer(self,):
        """
        VMDAS will record two buffers in NB or NB/BB mode, so we need to
        figure out if that is happening here
        """
        found = False
        pos = self.f.pos
        if self._debug_level >= 0:
            logging.info(self.hdr)
            logging.info('pos {}'.format(pos))
        self.id_positions = {}
        for offset in self.hdr['dat_offsets']:
            self.f.seek(offset+pos - self.hdr['dat_offsets'][0], rel=0)
            id = self.f.read_ui16(1)
            self.id_positions[id] = offset
            if self._debug_level >= 0:
                logging.info('pos {} id {}'.format(offset, id))
            if id == 1:
                self.read_fixed(bb=True)
                found = True
            elif id == 0:
                self.read_fixed(bb=False)
            elif id == 16:
                self.read_fixed_sl()  # bb=True
            elif id == 8192:
                self._vmdas_search = True
        return found

    def mean(self, dat):
        if self.n_avg == 1:
            return dat[..., 0]
        return np.nanmean(dat, axis=-1)

    def load_data(self, nens=None):
        if nens is None:
            self._nens = int(self._npings / self.n_avg)
        elif (nens.__class__ is tuple or nens.__class__ is list):
            raise Exception("    `nens` must be a integer")
        else:
            self._nens = nens
        if self._debug_level >= 0:
            logging.info('  taking data from pings 0 - %d' % self._nens)
            logging.info('  %d ensembles will be produced.\n' % self._nens)
        self.init_data()

        for iens in range(self._nens):
            if not self.read_buffer():
                self.remove_end(iens)
                break
            self.ensemble.clean_data()
            if self._bb:
                self.ensembleBB.clean_data()
            ens = [self.ensemble]
            vars = [self.vars_read]
            datl = [self.outd]
            if self._bb:
                ens += [self.ensembleBB]
                vars += [self.vars_readBB]
                datl += [self.outdBB]

            for var, en, dat in zip(vars, ens, datl):
                clock = en.rtc[:, :]
                if clock[0, 0] < 100:
                    clock[0, :] += defs.century

                for nm in var:
                    # If n_cells has increased (WinRiver transects)
                    ds = defs._get(dat, nm)
                    bn = self.mean(en[nm])
                    # Check that
                    # 1. n_cells has changed,
                    # 2. nm is a beam variable
                    # 3. n_cells is greater than any previous
                    if self.flag > 0 and len(ds.shape) == 3 and (ds.shape[0] != bn.shape[0]):
                        # increase the size of original dataset
                        a = np.empty(
                            (self.flag, ds.shape[1], ds.shape[2]))*np.nan
                        ds = np.append(ds, a, axis=0)
                        defs._setd(dat, nm, ds)
                    # Copy the ensemble to the dataset.
                    ds[..., iens] = bn
                # reset after all variables run
                self.flag = 0

                try:
                    dates = tmlib.date2epoch(
                        tmlib.datetime(*clock[:6, 0],
                                       microsecond=clock[6, 0] * 10000))[0]
                except ValueError:
                    warnings.warn("Invalid time stamp in ping {}.".format(
                        int(self.ensemble.number[0])))
                    dat['coords']['time'][iens] = np.NaN
                else:
                    dat['coords']['time'][iens] = np.median(dates)

        self.cleanup(self.cfg, self.outd)
        if self._bb:
            self.cleanup(self.cfgbb, self.outdBB)

        # Finalize dataset (runs through both nb and bb)
        for dat in datl:
            self.finalize(dat)
            if 'vel_bt' in dat['data_vars']:
                dat['attrs']['rotate_vars'].append('vel_bt')

        dat = self.outd
        datbb = self.outdBB if self._bb else None
        return dat, datbb

    def init_data(self,):
        outd = {'data_vars': {}, 'coords': {},
                'attrs': {}, 'units': {}, 'sys': {}}
        outd['attrs']['inst_make'] = 'TRDI'
        outd['attrs']['inst_type'] = 'ADCP'
        outd['attrs']['rotate_vars'] = ['vel', ]
        # Currently RDI doesn't use IMUs
        outd['attrs']['has_imu'] = 0
        if self._bb:
            outdbb = {'data_vars': {}, 'coords': {},
                      'attrs': {}, 'units': {}, 'sys': {}}
            outdbb['attrs']['inst_make'] = 'TRDI'
            outdbb['attrs']['inst_type'] = 'ADCP'
            outdbb['attrs']['rotate_vars'] = ['vel', ]
            outdbb['attrs']['has_imu'] = 0

        for nm in defs.data_defs:
            outd = defs._idata(outd, nm,
                               sz=defs._get_size(nm, self._nens, self.cfg['n_cells']))
        self.outd = outd

        if self._bb:
            for nm in defs.data_defs:
                outdbb = defs._idata(outdbb, nm,
                                     sz=defs._get_size(nm, self._nens, self.cfgbb['n_cells']))
            self.outdBB = outdbb
            if self._debug_level > 1:
                logging.info(np.shape(outdbb['data_vars']['vel']))

        if self._debug_level > 1:
            logging.info('{} ncells, not BB'.format(self.cfg['n_cells']))
            if self._bb:
                logging.info('{} ncells, BB'.format(self.cfgbb['n_cells']))

    def read_buffer(self,):
        fd = self.f
        self.ensemble.k = -1  # so that k+=1 gives 0 on the first loop.
        if self._bb:
            self.ensembleBB.k = -1  # so that k+=1 gives 0 on the first loop.
        self.print_progress()
        hdr = self.hdr
        while self.ensemble.k < self.ensemble.n_avg - 1:
            if not self.search_buffer():
                return False
            startpos = fd.tell() - 2
            self.read_hdrseg()
            if self._debug_level >= 0:
                logging.info('Read Header', hdr)
            byte_offset = self._nbyte + 2
            self._read_vmdas = False
            for n in range(len(hdr['dat_offsets'])):
                id = fd.read_ui16(1)
                if self._debug_level > 0:
                    logging.info(f'n {n}: {id} {id:04x}')
                self.print_pos()
                retval = self.read_dat(id)

                if retval == 'FAIL':
                    break
                byte_offset += self._nbyte
                if n < (len(hdr['dat_offsets']) - 1):
                    oset = hdr['dat_offsets'][n + 1] - byte_offset
                    if oset != 0:
                        if self._debug_level > 0:
                            logging.debug(
                                '  %s: Adjust location by %d\n' % (id, oset))
                        fd.seek(oset, 1)
                    byte_offset = hdr['dat_offsets'][n + 1]
                else:
                    if hdr['nbyte'] - 2 != byte_offset:
                        if not self._winrivprob:
                            if self._debug_level > 0:
                                logging.debug('  {:d}: Adjust location by {:d}\n'
                                              .format(id, hdr['nbyte'] - 2 - byte_offset))
                            self.f.seek(hdr['nbyte'] - 2 - byte_offset, 1)
                    byte_offset = hdr['nbyte'] - 2
            # Check for vmdas again because vmdas doesn't set the offsets
            # correctly, and we need this info:
            if not self._read_vmdas and self._vmdas_search:
                if self._debug_level >= 1:
                    logging.info(
                        'Searching for vmdas nav data. Going to next ensemble')
                self.search_buffer()
                # now go back to where vmdas would be:
                fd.seek(-98, 1)
                id = self.f.read_ui16(1)
                if id is not None:
                    if self._debug_level >= 1:
                        logging.info(f'Found {id:04d}')
                    if id == 8192:
                        self.read_dat(id)
            readbytes = fd.tell() - startpos
            offset = hdr['nbyte'] + 2 - readbytes
            self.check_offset(offset, readbytes)
            self.print_pos(byte_offset=byte_offset)

        return True

    def search_buffer(self):
        """
        Check to see if the next bytes indicate the beginning of a
        data block.  If not, search for the next data block, up to
        _search_num times.
        """
        id = self.f.read_ui8(2)
        if id is None:
            return False
        id1 = list(id)
        search_cnt = 0
        fd = self.f
        if self._debug_level >= 2:
            logging.info('  -->In search_buffer...')
        while (search_cnt < self._search_num and
               ((id1[0] != 127 or id1[1] != 127) or
                not self.checkheader())):
            search_cnt += 1
            nextbyte = fd.read_ui8(1)
            if nextbyte == None:
                return False
            id1[1] = id1[0]
            id1[0] = nextbyte
        if search_cnt == self._search_num:
            raise Exception(
                'Searched {} entries... Bad data encountered. -> {}'
                .format(search_cnt, id1))
        elif search_cnt > 0:
            if self._debug_level >= 1:
                logging.info('  Searched {} bytes to find next '
                             'valid ensemble start [{:x}, {:x}]\n'
                             .format(search_cnt, *id1))
        return True

    def checkheader(self,):
        if self._debug_level > 1:
            logging.info("  ###In checkheader.")
        fd = self.f
        valid = False
        if self._debug_level >= 0:
            logging.info('pos {}'.format(self.f.pos))
        numbytes = fd.read_i16(1)
        if numbytes > 0:
            fd.seek(numbytes - 2, 1)
            cfgid = fd.read_ui8(2)
            if cfgid is None:
                if self._debug_level > 1:
                    logging.info('EOF')
                return False
            if len(cfgid) == 2:
                fd.seek(-numbytes - 2, 1)
                if cfgid[0] == 127 and cfgid[1] in [127, 121]:
                    if cfgid[1] == 121 and self._debug7f79 is None:
                        self._debug7f79 = True
                        if self._debug_level > 1:
                            logging.warning('7f79!!!')
                    valid = True
        else:
            fd.seek(-2, 1)
        if self._debug_level > 1:
            logging.info("  ###Leaving checkheader.")
        return valid

    def read_hdrseg(self,):
        fd = self.f
        hdr = self.hdr
        hdr['nbyte'] = fd.read_i16(1)
        spare = fd.read_ui8(1)
        ndat = fd.read_ui8(1)
        hdr['dat_offsets'] = fd.read_ui16(ndat)
        self._nbyte = 4 + ndat * 2

    def print_progress(self,):
        self.progress = self.f.tell()
        if self._debug_level > 1:
            logging.debug('  pos %0.0fmb/%0.0fmb\n' %
                          (self.f.tell() / 1048576., self._filesize / 1048576.))
        if (self.f.tell() - self.progress) < 1048576:
            return

    def print_pos(self, byte_offset=-1):
        """Print the position in the file, used for debugging.
        """
        if self._debug_level >= 2:
            if hasattr(self, 'ensemble'):
                k = self.ensemble.k
            else:
                k = 0
            logging.debug(
                f'  pos: {self.f.tell()}, pos_: {self._pos}, nbyte: {self._nbyte}, k: {k}, byte_offset: {byte_offset}')

    def check_offset(self, offset, readbytes):
        fd = self.f
        if offset != 4 and self._fixoffset == 0:
            if self._debug_level > 0:
                if fd.tell() == self._filesize:
                    logging.error(
                        ' EOF reached unexpectedly - discarding this last ensemble\n')
                else:
                    logging.debug("  Adjust location by {:d} (readbytes={:d},hdr['nbyte']={:d})\n"
                                  .format(offset, readbytes, self.hdr['nbyte']))
            self._fixoffset = offset - 4
        fd.seek(4 + self._fixoffset, 1)

    def remove_end(self, iens):
        dat = self.outd
        if self._debug_level > 0:
            logging.info('  Encountered end of file.  Cleaning up data.')
        for nm in self.vars_read:
            defs._setd(dat, nm, defs._get(dat, nm)[..., :iens])

    def read_dat(self, id):
        function_map = {0: (self.read_fixed, []),   # 0000 1st profile fixed leader
                        1:  (self.read_fixed, [True]),  # 0001
                        # 0010 Surface layer fixed leader (RiverPro & StreamPro)
                        16: (self.read_fixed_sl, []),
                        # 0080 1st profile variable leader
                        128: (self.read_var, [0]),
                        # 0081 2nd profile variable leader
                        129: (self.read_var, [1]),
                        # 0100 1st profile velocity
                        256: (self.read_vel, [0]),
                        # 0101 2nd profile velocity
                        257: (self.read_vel, [1]),
                        # 0103 Waves first leader
                        259: (self.skip_Nbyte, [74]),
                        # 0110 Surface layer velocity (RiverPro & StreamPro)
                        272: (self.read_vel, [2]),
                        # 0200 1st profile correlation
                        512: (self.read_corr, [0]),
                        # 0201 2nd profile correlation
                        513: (self.read_corr, [1]),
                        # 0203 Waves data
                        515: (self.skip_Nbyte, [186]),
                        # 020C Ambient sound profile
                        524: (self.skip_Nbyte, [4]),
                        # 0210 Surface layer correlation (RiverPro & StreamPro)
                        528: (self.read_corr, [2]),
                        # 0300 1st profile amplitude
                        768: (self.read_amp, [0]),
                        # 0301 2nd profile amplitude
                        769: (self.read_amp, [1]),
                        # 0302 Beam 5 Sum of squared velocities
                        770: (self.skip_Ncol, []),
                        # 0303 Waves last leader
                        771: (self.skip_Ncol, [18]),
                        # 0310 Surface layer amplitude (RiverPro & StreamPro)
                        784: (self.read_amp, [2]),
                        # 0400 1st profile % good
                        1024: (self.read_prcnt_gd, [0]),
                        # 0401 2nd profile pct good
                        1025: (self.read_prcnt_gd, [1]),
                        # 0403 Waves HPR data
                        1027: (self.skip_Nbyte, [6]),
                        # 0410 Surface layer pct good (RiverPro & StreamPro)
                        1040: (self.read_prcnt_gd, [2]),
                        # 0500 1st profile status
                        1280: (self.read_status, [0]),
                        # 0501 2nd profile status
                        1281: (self.read_status, [1]),
                        # 0510 Surface layer status (RiverPro & StreamPro)
                        1296: (self.read_status, [2]),
                        1536: (self.read_bottom, []),  # 0600 bottom tracking
                        1793: (self.skip_Ncol, [4]),  # 0701 number of pings
                        1794: (self.skip_Ncol, [4]),  # 0702 sum of squared vel
                        1795: (self.skip_Ncol, [4]),  # 0703 sum of velocities
                        2560: (self.skip_Ncol, []),  # 0A00 Beam 5 velocity
                        2816: (self.skip_Ncol, []),  # 0B00 Beam 5 correlation
                        3072: (self.skip_Ncol, []),  # 0C00 Beam 5 amplitude
                        3328: (self.skip_Ncol, []),  # 0D00 Beam 5 pct_good
                        # Fixed attitude data format for Ocean Surveyor ADCPs
                        3000: (self.skip_Nbyte, [32]),
                        3841: (self.skip_Nbyte, [38]),  # 0F01 Beam 5 leader
                        8192: (self.read_vmdas, []),   # 2000
                        # 2013 Navigation parameter data
                        8211: (self.skip_Nbyte, [83]),
                        8226: (self.read_winriver2, []),  # 2022
                        8448: (self.read_winriver, [38]),  # 2100
                        8449: (self.read_winriver, [97]),  # 2101
                        8450: (self.read_winriver, [45]),  # 2102
                        8451: (self.read_winriver, [60]),  # 2103
                        8452: (self.read_winriver, [38]),  # 2104
                        # 3200 Transformation matrix
                        12800: (self.skip_Nbyte, [32]),
                        # 3000 Fixed attitude data format for Ocean Surveyor ADCPs
                        12288: (self.skip_Nbyte, [32]),
                        # 4100 beam 5 range
                        16640: (self.read_alt, []),
                        # 4400 Firmware status data (RiverPro & StreamPro)
                        17408: (self.skip_Nbyte, [28]),
                        # 4401 Auto mode setup (RiverPro & StreamPro)
                        17409: (self.skip_Nbyte, [82]),
                        # 5803 High resolution bottom track velocity
                        22531: (self.skip_Nbyte, [68]),
                        # 5804 Bottom track range
                        22532: (self.skip_Nbyte, [21]),
                        # 5901 ISM (IMU) data
                        22785: (self.skip_Nbyte, [65]),
                        # 5902 Ping attitude
                        22786: (self.skip_Nbyte, [105]),
                        # 7001 ADC data
                        28673: (self.skip_Nbyte, [14]),
                        }
        # Call the correct function:
        if self._debug_level >= 2:
            logging.debug(f'Trying to Read {id}')
        if id in function_map:
            if self._debug_level > 1:
                logging.info('  Reading code {}...'.format(hex(id)))
            retval = function_map.get(id)[0](*function_map[id][1])
            if retval:
                return retval
            if self._debug_level > 1:
                logging.info('    success!')
        else:
            self.read_nocode(id)

    def read_fixed(self, bb=False):
        self.read_cfgseg(bb=bb)
        self._nbyte += 2
        if self._debug_level >= 0:
            logging.info('Read Fixed')

        # Check if n_cells changed (for winriver transect files)
        if hasattr(self, 'ensemble') and (self.ensemble['n_cells'] != self.cfg['n_cells']):
            diff = self.cfg['n_cells'] - self.ensemble['n_cells']
            if diff > 0:
                self.flag = diff
                self.ensemble = defs._ensemble(self.n_avg, self.cfg['n_cells'])
                # Not concerned if # of cells decreases
                if self._debug_level >= 1:
                    logging.warning('Number of cells changed to {}'
                                    .format(self.cfg['n_cells']))

    def read_fixed_sl(self,):
        # Surface layer profile
        cfg = self.cfg
        cfg['surface_layer'] = 1
        cfg['n_cells_sl'] = self.f.read_ui8(1)
        cfg['cell_size_sl'] = self.f.read_ui16(1) * .01
        cfg['bin1_dist_m_sl'] = round(self.f.read_ui16(1) * .01, 4)

        if self._debug_level >= 0:
            logging.info('Read Surface Layer Config')
        self._nbyte = 2 + 5

    def read_cfgseg(self, bb=False):
        cfgstart = self.f.tell()

        if bb:
            cfg = self.cfgbb
        else:
            cfg = self.cfg
        fd = self.f
        tmp = fd.read_ui8(5)
        prog_ver0 = tmp[0]
        cfg['prog_ver'] = tmp[0] + tmp[1] / 100.
        cfg['inst_model'] = defs.adcp_type.get(tmp[0],
                                               'unrecognized firmware version')
        config = tmp[2:4]
        cfg['beam_angle'] = [15, 20, 30][(config[1] & 3)]
        beam5 = [0, 1][int((config[1] & 16) == 16)]
        cfg['freq'] = ([75, 150, 300, 600, 1200, 2400, 38][(config[0] & 7)])
        cfg['beam_pattern'] = (['concave',
                                'convex'][int((config[0] & 8) == 8)])
        cfg['orientation'] = ['down', 'up'][int((config[0] & 128) == 128)]
        simflag = ['real', 'simulated'][tmp[4]]
        fd.seek(1, 1)
        cfg['n_beams'] = fd.read_ui8(1) + beam5
        cfg['n_cells'] = fd.read_ui8(1)
        cfg['pings_per_ensemble'] = fd.read_ui16(1)
        cfg['cell_size'] = fd.read_ui16(1) * .01
        cfg['blank_dist'] = fd.read_ui16(1) * .01
        cfg['profiling_mode'] = fd.read_ui8(1)
        cfg['min_corr_threshold'] = fd.read_ui8(1)
        cfg['n_code_reps'] = fd.read_ui8(1)
        cfg['min_prcnt_gd'] = fd.read_ui8(1)
        cfg['max_error_vel'] = fd.read_ui16(1) / 1000
        cfg['sec_between_ping_groups'] = (
            np.sum(np.array(fd.read_ui8(3)) *
                   np.array([60., 1., .01])))
        coord_sys = fd.read_ui8(1)
        cfg['coord_sys'] = (['beam', 'inst',
                             'ship', 'earth'][((coord_sys >> 3) & 3)])
        cfg['use_pitchroll'] = ['no', 'yes'][(coord_sys & 4) == 4]
        cfg['use_3beam'] = ['no', 'yes'][(coord_sys & 2) == 2]
        cfg['bin_mapping'] = ['no', 'yes'][(coord_sys & 1) == 1]
        cfg['heading_misalign_deg'] = fd.read_i16(1) * .01
        cfg['magnetic_var_deg'] = fd.read_i16(1) * .01
        cfg['sensors_src'] = np.binary_repr(fd.read_ui8(1), 8)
        cfg['sensors_avail'] = np.binary_repr(fd.read_ui8(1), 8)
        cfg['bin1_dist_m'] = round(fd.read_ui16(1) * .01, 4)
        cfg['transmit_pulse_m'] = fd.read_ui16(1) * .01
        cfg['water_ref_cells'] = list(fd.read_ui8(2))  # list for attrs
        cfg['false_target_threshold'] = fd.read_ui8(1)
        fd.seek(1, 1)
        cfg['transmit_lag_m'] = fd.read_ui16(1) * .01
        self._nbyte = 40

        if cfg['prog_ver'] >= 8.14:
            cpu_serialnum = fd.read_ui8(8)
            self._nbyte += 8
        if cfg['prog_ver'] >= 8.24:
            cfg['bandwidth'] = fd.read_ui16(1)
            self._nbyte += 2
        if cfg['prog_ver'] >= 16.05:
            cfg['power_level'] = fd.read_ui8(1)
            self._nbyte += 1
        if cfg['prog_ver'] >= 16.27:
            # cfg['navigator_basefreqindex'] = fd.read_ui8(1)
            fd.seek(1, 1)
            cfg['serialnum'] = fd.read_ui32(1)
            cfg['beam_angle'] = fd.read_ui8(1)
            self._nbyte += 6
        # # Documented in latest PDDecoder 10/12/22 but must be instrument specific
        # if cfg['prog_ver'] >= 56:
        #     fd.seek(1, 1)
        #     pings_per_ensemble = fd.read_ui16(1)
        #     exact_freq = fd.read_ui8(3)
        #     #cfg['exact_freq'] = int("".join(str(x) for x in exact_freq))
        #     self._nbyte += 6

        self.configsize = self.f.tell() - cfgstart
        if self._debug_level >= 0:
            logging.info('Read Config')

    def read_var(self, bb=False):
        """ Read variable leader """
        fd = self.f
        if bb:
            ens = self.ensembleBB
        else:
            ens = self.ensemble
        ens.k += 1
        ens = self.ensemble
        k = ens.k
        self.vars_read += ['number',
                           'rtc',
                           'number',
                           'builtin_test_fail',
                           'c_sound',
                           'depth',
                           'heading',
                           'pitch',
                           'roll',
                           'salinity',
                           'temp',
                           'min_preping_wait',
                           'heading_std',
                           'pitch_std',
                           'roll_std',
                           'adc']
        ens.number[k] = fd.read_ui16(1)
        ens.rtc[:, k] = fd.read_ui8(7)
        ens.number[k] += 65535 * fd.read_ui8(1)
        ens.builtin_test_fail[k] = fd.read_ui16(1)
        ens.c_sound[k] = fd.read_ui16(1)
        ens.depth[k] = fd.read_ui16(1) * 0.1
        ens.heading[k] = fd.read_ui16(1) * 0.01
        ens.pitch[k] = fd.read_i16(1) * 0.01
        ens.roll[k] = fd.read_i16(1) * 0.01
        ens.salinity[k] = fd.read_i16(1)
        ens.temp[k] = fd.read_i16(1) * 0.01
        ens.min_preping_wait[k] = (fd.read_ui8(
            3) * np.array([60, 1, .01])).sum()
        ens.heading_std[k] = fd.read_ui8(1)
        ens.pitch_std[k] = fd.read_ui8(1) * 0.1
        ens.roll_std[k] = fd.read_ui8(1) * 0.1
        ens.adc[:, k] = fd.read_i8(8)
        self._nbyte = 2 + 40

        cfg = self.cfg
        if cfg['inst_model'].lower() == 'broadband':
            if cfg['prog_ver'] >= 5.55:
                fd.seek(15, 1)
                cent = fd.read_ui8(1)
                ens.rtc[:, k] = fd.read_ui8(7)
                ens.rtc[0, k] = ens.rtc[0, k] + cent * 100
                self._nbyte += 23
        elif cfg['inst_model'].lower() == 'ocean surveyor':
            fd.seek(16, 1)  # 30 bytes all set to zero, 14 read above
            self._nbyte += 16
            if cfg['prog_ver'] > 23:
                fd.seek(2, 1)
                self._nbyte += 2
        else:
            ens.error_status[k] = np.binary_repr(fd.read_ui32(1), 32)
            self.vars_read += ['pressure', 'pressure_std']
            self._nbyte += 4
            if cfg['prog_ver'] >= 8.13:
                # Added pressure sensor stuff in 8.13
                fd.seek(2, 1)
                ens.pressure[k] = fd.read_ui32(1) / 1000  # dPa to dbar
                ens.pressure_std[k] = fd.read_ui32(1) / 1000
                self._nbyte += 10
            if cfg['prog_ver'] >= 8.24:
                # Spare byte added 8.24
                fd.seek(1, 1)
                self._nbyte += 1
            if cfg['prog_ver'] >= 16.05:
                # Added more fields with century in clock
                cent = fd.read_ui8(1)
                ens.rtc[:, k] = fd.read_ui8(7)
                ens.rtc[0, k] = ens.rtc[0, k] + cent * 100
                self._nbyte += 8
            if cfg['prog_ver'] >= 56:
                fd.seek(1)  # lag near bottom flag
                self._nbyte += 1

        if self._debug_level >= 0:
            logging.info('Read Var')

    def switch_profile(self, bb):
        if bb == 1:
            ens = self.ensembleBB
            cfg = self.cfgbb
            # Placeholder for dual profile mode
            # Solution for vmdas profile in bb spot (vs nb)
            tag = ''
        elif bb == 2:
            ens = self.ensemble
            cfg = self.cfg
            tag = '_sl'
        else:
            ens = self.ensemble
            cfg = self.cfg
            tag = ''

        return ens, cfg, tag

    def read_vel(self, bb=0):
        ens, cfg, tg = self.switch_profile(bb)
        self.vars_read += ['vel'+tg]
        n_cells = cfg['n_cells'+tg]

        k = ens.k
        vel = np.array(
            self.f.read_i16(4 * n_cells)
        ).reshape((n_cells, 4)) * .001
        ens['vel'+tg][:n_cells, :, k] = vel
        self._nbyte = 2 + 4 * n_cells * 2
        if self._debug_level >= 0:
            logging.info('Read Vel')

    def read_corr(self, bb=0):
        ens, cfg, tg = self.switch_profile(bb)
        self.vars_read += ['corr'+tg]
        n_cells = cfg['n_cells'+tg]

        k = ens.k
        ens['corr'+tg][:n_cells, :, k] = np.array(
            self.f.read_ui8(4 * n_cells)
        ).reshape((n_cells, 4))
        self._nbyte = 2 + 4 * n_cells
        if self._debug_level >= 0:
            logging.info('Read Corr')

    def read_amp(self, bb=0):
        ens, cfg, tg = self.switch_profile(bb)
        self.vars_read += ['amp'+tg]
        n_cells = cfg['n_cells'+tg]

        k = ens.k
        ens['amp'+tg][:n_cells, :, k] = np.array(
            self.f.read_ui8(4 * n_cells)
        ).reshape((n_cells, 4))
        self._nbyte = 2 + 4 * n_cells
        if self._debug_level >= 0:
            logging.info('Read Amp')

    def read_prcnt_gd(self, bb=0):
        ens, cfg, tg = self.switch_profile(bb)
        self.vars_read += ['prcnt_gd'+tg]
        n_cells = cfg['n_cells'+tg]

        ens['prcnt_gd'+tg][:n_cells, :, ens.k] = np.array(
            self.f.read_ui8(4 * n_cells)
        ).reshape((n_cells, 4))
        self._nbyte = 2 + 4 * n_cells
        if self._debug_level >= 0:
            logging.info('Read PG')

    def read_status(self, bb=0):
        ens, cfg, tg = self.switch_profile(bb)
        self.vars_read += ['status'+tg]
        n_cells = cfg['n_cells'+tg]

        ens['status'+tg][:n_cells, :, ens.k] = np.array(
            self.f.read_ui8(4 * n_cells)
        ).reshape((n_cells, 4))
        self._nbyte = 2 + 4 * n_cells
        if self._debug_level >= 0:
            logging.info('Read Status')

    def read_bottom(self,):
        self.vars_read += ['dist_bt', 'vel_bt', 'corr_bt', 'amp_bt',
                           'prcnt_gd_bt']
        fd = self.f
        ens = self.ensemble
        k = ens.k
        cfg = self.cfg
        if self._source == 2:
            self.vars_read += ['latitude_gps', 'longitude_gps']
            fd.seek(2, 1)
            long1 = fd.read_ui16(1)
            fd.seek(6, 1)
            ens.latitude_gps[k] = fd.read_i32(1) * self._cfac
            if ens.latitude_gps[k] == 0:
                ens.latitude_gps[k] = np.NaN
        else:
            fd.seek(14, 1)
        ens.dist_bt[:, k] = fd.read_ui16(4) * 0.01
        ens.vel_bt[:, k] = fd.read_i16(4) * 0.001
        ens.corr_bt[:, k] = fd.read_ui8(4)
        ens.amp_bt[:, k] = fd.read_ui8(4)
        ens.prcnt_gd_bt[:, k] = fd.read_ui8(4)
        if self._source == 2:
            fd.seek(2, 1)
            ens.longitude_gps[k] = (
                long1 + 65536 * fd.read_ui16(1)) * self._cfac
            if ens.longitude_gps[k] > 180:
                ens.longitude_gps[k] = ens.longitude_gps[k] - 360
            if ens.longitude_gps[k] == 0:
                ens.longitude_gps[k] = np.NaN
            fd.seek(16, 1)
            qual = fd.read_ui8(1)
            if qual == 0:
                if self._debug_level > 0:
                    logging.info('  qual==%d,%f %f' % (qual,
                                                       ens.latitude_gps[k],
                                                       ens.longitude_gps[k]))
                ens.latitude_gps[k] = np.NaN
                ens.longitude_gps[k] = np.NaN
            fd.seek(71 - 45 - 16 - 17, 1)
            self._nbyte = 2 + 68
        else:
            # Skip reference layer data
            fd.seek(26, 1)
            self._nbyte = 2 + 68
        if cfg['prog_ver'] >= 5.3:
            fd.seek(7, 1)  # skip to rangeMsb bytes
            ens.dist_bt[:, k] = ens.dist_bt[:, k] + fd.read_ui8(4) * 655.36
            self._nbyte += 11
        if cfg['prog_ver'] >= 16.2 and (cfg.get('sourceprog') != 'WINRIVER'):
            fd.seek(4, 1)  # not documented
            self._nbyte += 4
        if cfg['prog_ver'] >= 56.1:
            fd.seek(4, 1)  # not documented
            self._nbyte += 4

        if self._debug_level >= 0:
            logging.info('Read Bottom Track')

    def read_alt(self,):
        """Read altimeter (vertical beam range) """
        fd = self.f
        ens = self.ensemble
        k = ens.k
        self.vars_read += ['alt_dist', 'alt_rssi', 'alt_eval', 'alt_status']
        ens.alt_eval[k] = fd.read_ui8(1)  # evaluation amplitude
        ens.alt_rssi[k] = fd.read_ui8(1)  # RSSI amplitude
        ens.alt_dist[k] = fd.read_ui32(1) / 1000  # range to surface/seafloor
        ens.alt_status[k] = fd.read_ui8(1)  # status bit flags
        self._nbyte = 7 + 2
        if self._debug_level >= 0:
            logging.info('Read Altimeter')

    def read_vmdas(self,):
        """Read VMDAS Navigation block"""
        fd = self.f
        self.cfg['sourceprog'] = 'VMDAS'
        ens = self.ensemble
        k = ens.k
        if self._source != 1 and self._debug_level >= 0:
            logging.info('  \n***** Apparently a VMDAS file \n\n')
        self._source = 1
        self.vars_read += ['time_gps',
                           'clock_offset_UTC_gps',
                           'latitude_gps',
                           'longitude_gps',
                           'avg_speed_gps',
                           'avg_dir_gps',
                           'speed_made_good_gps',
                           'dir_made_good_gps',
                           'flags_gps',
                           ]
        # UTC date time
        utim = fd.read_ui8(4)
        date_utc = tmlib.datetime(utim[2] + utim[3] * 256, utim[1], utim[0])

        # 1st lat/lon position after previous ADCP ping
        # This byte is in hundredths of seconds (10s of milliseconds):
        utc_time_first_fix = tmlib.timedelta(
            milliseconds=(int(fd.read_ui32(1) / 10)))
        ens.clock_offset_UTC_gps[k] = fd.read_i32(
            1) / 1000  # "PC clock offset from UTC" in ms
        latitude_first_gps = fd.read_i32(1) * self._cfac
        longitude_first_gps = fd.read_i32(1) * self._cfac

        # Last lat/lon position prior to current ADCP ping
        utc_time_fix = tmlib.timedelta(
            milliseconds=(int(fd.read_ui32(1) / 10)))
        ens.time_gps[k] = tmlib.date2epoch(date_utc + utc_time_fix)[0]
        ens.latitude_gps[k] = fd.read_i32(1) * self._cfac
        ens.longitude_gps[k] = fd.read_i32(1) * self._cfac

        ens.avg_speed_gps[k] = fd.read_ui16(1) / 1000
        ens.avg_dir_gps[k] = fd.read_ui16(1) * 180 / 2 ** 15  # avg true track
        fd.seek(2, 1)  # avg magnetic track
        ens.speed_made_good_gps[k] = fd.read_ui16(1) / 1000
        ens.dir_made_good_gps[k] = fd.read_ui16(1) * 180 / 2 ** 15
        fd.seek(2, 1)  # reserved
        ens.flags_gps[k] = int(np.binary_repr(fd.read_ui16(1)))
        fd.seek(6, 1)  # reserved, ADCP ensemble #

        # ADCP date time
        utim = fd.read_ui8(4)
        date_adcp = tmlib.datetime(utim[0] + utim[1] * 256, utim[3], utim[2])
        time_adcp = tmlib.timedelta(
            milliseconds=(int(fd.read_ui32(1) / 10)))

        fd.seek(16, 1)
        self._nbyte = 2 + 76

        if self._debug_level >= 0:
            logging.info('Read VMDAS')
        self._read_vmdas = True

    def read_winriver2(self, ):
        startpos = self.f.tell()
        self._winrivprob = True
        self.cfg['sourceprog'] = 'WinRiver2'
        ens = self.ensemble
        k = ens.k
        if self._debug_level >= 0:
            logging.info('Read WinRiver2')
        self._source = 3

        spid = self.f.read_ui16(1)  # NMEA specific IDs
        if spid in [4, 104]:  # GGA
            sz = self.f.read_ui16(1)
            dtime = self.f.read_f64(1)
            if sz <= 43:  # If no sentence, data is still stored in nmea format
                empty_gps = self.f.reads(sz-2)
                self.f.seek(2, 1)
            else:  # TRDI rewrites the nmea string into their format if one is found
                start_string = self.f.reads(6)
                if type(start_string) != str:
                    if self._debug_level >= 1:
                        logging.warning(f'Invalid GGA string found in ensemble {k},'
                                        ' skipping...')
                    return 'FAIL'
                self.f.seek(1, 1)
                gga_time = self.f.reads(9)
                time = tmlib.timedelta(hours=int(gga_time[0:2]),
                                       minutes=int(gga_time[2:4]),
                                       seconds=int(gga_time[4:6]),
                                       milliseconds=int(float(gga_time[6:])*1000))
                clock = self.ensemble.rtc[:, :]
                if clock[0, 0] < 100:
                    clock[0, :] += defs.century
                date = tmlib.datetime(*clock[:3, 0]) + time
                ens.time_gps[k] = tmlib.date2epoch(date)[0]
                self.f.seek(1, 1)
                ens.latitude_gps[k] = self.f.read_f64(1)
                tcNS = self.f.reads(1)  # 'N' or 'S'
                if tcNS == 'S':
                    ens.latitude_gps[k] *= -1
                ens.longitude_gps[k] = self.f.read_f64(1)
                tcEW = self.f.reads(1)  # 'E' or 'W'
                if tcEW == 'W':
                    ens.longitude_gps[k] *= -1
                ens.fix_gps[k] = self.f.read_ui8(1)  # gps fix type/quality
                ens.n_sat_gps[k] = self.f.read_ui8(1)  # of satellites
                # horizontal dilution of precision
                ens.hdop_gps[k] = self.f.read_float(1)
                ens.elevation_gps[k] = self.f.read_float(1)  # altitude
                m = self.f.reads(1)  # altitude unit, 'm'
                h_geoid = self.f.read_float(1)  # height of geoid
                m2 = self.f.reads(1)  # geoid unit, 'm'
                ens.rtk_age_gps[k] = self.f.read_float(1)
                station_id = self.f.read_ui16(1)
            self.vars_read += ['time_gps', 'longitude_gps', 'latitude_gps', 'fix_gps',
                               'n_sat_gps', 'hdop_gps', 'elevation_gps', 'rtk_age_gps']
            self._nbyte = self.f.tell() - startpos + 2

        elif spid in [5, 105]:  # VTG
            sz = self.f.read_ui16(1)
            dtime = self.f.read_f64(1)
            if sz <= 22:  # if no data
                empty_gps = self.f.reads(sz-2)
                self.f.seek(2, 1)
            else:
                start_string = self.f.reads(6)
                if type(start_string) != str:
                    if self._debug_level >= 1:
                        logging.warning(f'Invalid VTG string found in ensemble {k},'
                                        ' skipping...')
                    return 'FAIL'
                self.f.seek(1, 1)
                true_track = self.f.read_float(1)
                t = self.f.reads(1)  # 'T'
                magn_track = self.f.read_float(1)
                m = self.f.reads(1)  # 'M'
                speed_knot = self.f.read_float(1)
                kts = self.f.reads(1)  # 'N'
                speed_kph = self.f.read_float(1)
                kph = self.f.reads(1)  # 'K'
                mode = self.f.reads(1)
                # knots -> m/s
                ens.speed_over_grnd_gps[k] = speed_knot / 1.944
                ens.dir_over_grnd_gps[k] = true_track
            self.vars_read += ['speed_over_grnd_gps',
                               'dir_over_grnd_gps']
            self._nbyte = self.f.tell() - startpos + 2

        elif spid in [6, 106]:  # 'DBT' depth sounder
            sz = self.f.read_ui16(1)
            dtime = self.f.read_f64(1)
            if sz <= 20:
                empty_gps = self.f.reads(sz-2)
                self.f.seek(2, 1)
            else:
                start_string = self.f.reads(6)
                if type(start_string) != str:
                    if self._debug_level >= 1:
                        logging.warning(f'Invalid DBT string found in ensemble {k},'
                                        ' skipping...')
                    return 'FAIL'
                self.f.seek(1, 1)
                depth_ft = self.f.read_float(1)
                ft = self.f.reads(1)  # 'f'
                depth_m = self.f.read_float(1)
                m = self.f.reads(1)  # 'm'
                depth_fathom = self.f.read_float(1)
                f = self.f.reads(1)  # 'F'
                ens.dist_nmea[k] = depth_m
            self.vars_read += ['dist_nmea']
            self._nbyte = self.f.tell() - startpos + 2

        elif spid in [7, 107]:  # 'HDT'
            sz = self.f.read_ui16(1)
            dtime = self.f.read_f64(1)
            if sz <= 14:
                empty_gps = self.f.reads(sz-2)
                self.f.seek(2, 1)
            else:
                start_string = self.f.reads(6)
                if type(start_string) != str:
                    if self._debug_level >= 1:
                        logging.warning(f'Invalid HDT string found in ensemble {k},'
                                        ' skipping...')
                    return 'FAIL'
                self.f.seek(1, 1)
                ens.heading_gps[k] = self.f.read_f64(1)
                tt = self.f.reads(1)
            self.vars_read += ['heading_gps']
            self._nbyte = self.f.tell() - startpos + 2

    def read_winriver(self, nbt):
        self._winrivprob = True
        self.cfg['sourceprog'] = 'WINRIVER'
        if self._source not in [2, 3]:
            if self._debug_level >= 0:
                logging.warning('\n***** Apparently a WINRIVER file - '
                                'Raw NMEA data handler not yet implemented\n')
            self._source = 2
        startpos = self.f.tell()
        sz = self.f.read_ui16(1)
        tmp = self.f.reads(sz-2)
        self._nbyte = self.f.tell() - startpos + 2

    def skip_Ncol(self, n_skip=1):
        self.f.seek(n_skip * self.cfg['n_cells'], 1)
        self._nbyte = 2 + n_skip * self.cfg['n_cells']

    def skip_Nbyte(self, n_skip):
        self.f.seek(n_skip, 1)
        self._nbyte = 2 + n_skip

    def read_nocode(self, id):
        # Skipping bytes from codes 0340-30FC, commented if needed
        hxid = hex(id)
        if hxid[2:4] == '30':
            logging.warning("Skipping bytes from codes 0340-30FC")
            # I want to count the number of 1s in the middle 4 bits
            # of the 2nd two bytes.
            # 60 is a 0b00111100 mask
            nflds = (bin(int(hxid[3]) & 60).count('1') +
                     bin(int(hxid[4]) & 60).count('1'))
            # I want to count the number of 1s in the highest
            # 2 bits of byte 3
            # 3 is a 0b00000011 mask:
            dfac = bin(int(hxid[3], 0) & 3).count('1')
            self.skip_Nbyte(12 * nflds * dfac)
        else:
            if self._debug_level >= 0:
                logging.warning('  Unrecognized ID code: %0.4X' % id)
            self.skip_nocode(id)

    def skip_nocode(self, id):
        # Skipping bytes if ID isn't known
        offsets = list(self.id_positions.values())
        idx = np.where(offsets == self.id_positions[id])[0][0]
        byte_len = offsets[idx+1] - offsets[idx] - 2

        self.skip_Nbyte(byte_len)
        if self._debug_level >= 0:
            logging.debug(f"Skipping ID code {id}\n")

    def cleanup(self, cfg, dat):
        dat['coords']['range'] = (cfg['bin1_dist_m'] +
                                  np.arange(self.ensemble['n_cells']) *
                                  cfg['cell_size'])

        for nm in cfg:
            dat['attrs'][nm] = cfg[nm]

        if 'surface_layer' in cfg:  # RiverPro/StreamPro
            dat['coords']['range_sl'] = (cfg['bin1_dist_m_sl'] +
                                         np.arange(self.cfg['n_cells_sl']) *
                                         cfg['cell_size_sl'])
            # Trim surface layer profile to length
            dv = dat['data_vars']
            for var in dv:
                if 'sl' in var:
                    dv[var] = dv[var][:cfg['n_cells_sl']]
            dat['attrs']['rotate_vars'].append('vel_sl')

    def finalize(self, dat):
        """Remove the attributes from the data that were never loaded.
        """
        for nm in set(defs.data_defs.keys()) - self.vars_read:
            defs._pop(dat, nm)
        for nm in self.cfg:
            dat['attrs'][nm] = self.cfg[nm]

        # VMDAS and WinRiver have different set sampling frequency
        da = dat['attrs']
        if hasattr(da, 'sourceprog') and (da['sourceprog'].lower() in ['vmdas', 'winriver', 'winriver2']):
            da['fs'] = round(np.diff(dat['coords']['time']).mean() ** -1, 2)
        else:
            da['fs'] = (da['sec_between_ping_groups'] *
                        da['pings_per_ensemble']) ** (-1)
        da['n_cells'] = self.ensemble['n_cells']

        for nm in defs.data_defs:
            shp = defs.data_defs[nm][0]
            if len(shp) and shp[0] == 'nc' and defs._in_group(dat, nm):
                defs._setd(dat, nm, np.swapaxes(defs._get(dat, nm), 0, 1))

    def __enter__(self,):
        return self

    def __exit__(self, type, value, traceback):
        self.f.close()



#!/usr/bin/python
import argparse
from dolfyn import read, save_mat

parser = argparse.ArgumentParser(
    description="Converts acoustic Doppler instrument data from binary"
    " format to Matlab(TM) (.mat).")

parser.add_argument('files',
                    help="The filename(s) to convert.",
                    action='append',
                    )

args = parser.parse_args()

for fnm in args.files:
    dat = read(fnm)

    if '.' in fnm:
        ext = fnm[fnm.index('.'):]
    outnm = fnm.rstrip(ext) + '.mat'

    print('Saving to %s.' % outnm)
    save_mat(dat, outnm, datenum=True)



import numpy as np

century = 2000
adcp_type = {4: 'Broadband',
             5: 'Broadband',
             6: 'Navigator',
             10: 'Rio Grande',
             11: 'H-ADCP',
             14: 'Ocean Surveyor',
             16: 'Workhorse',
             19: 'Navigator',
             23: 'Ocean Surveyor',
             28: 'ChannelMaster',
             31: 'StreamPro',
             34: 'Explorer',
             37: 'Navigator',
             41: 'DVS',
             43: 'Workhorse',
             44: 'RiverRay',
             47: 'SentinelV',
             50: 'Workhorse',
             51: 'Workhorse',
             52: 'Workhorse',
             53: 'Navigator',
             55: 'DVS',
             56: 'RiverPro',
             59: 'Meridian',
             61: 'Pinnacle',
             66: 'SentinelV',
             67: 'Pathfinder',
             73: 'Pioneer',
             74: 'Tasman',
             76: 'WayFinder',
             77: 'Workhorse',
             78: 'Workhorse',
             }

data_defs = {'number': ([], 'data_vars', 'uint32', ''),
             'rtc': ([7], 'sys', 'uint16', ''),
             'builtin_test_fail': ([], 'data_vars', 'bool', ''),
             'c_sound': ([], 'data_vars', 'float32', 'm/s'),
             'depth': ([], 'data_vars', 'float32', 'm'),
             'pitch': ([], 'data_vars', 'float32', 'deg'),
             'roll': ([], 'data_vars', 'float32', 'deg'),
             'heading': ([], 'data_vars', 'float32', 'deg'),
             'temp': ([], 'data_vars', 'float32', 'C'),
             'salinity': ([], 'data_vars', 'float32', 'psu'),
             'min_preping_wait': ([], 'data_vars', 'float32', 's'),
             'heading_std': ([], 'data_vars', 'float32', 'deg'),
             'pitch_std': ([], 'data_vars', 'float32', 'deg'),
             'roll_std': ([], 'data_vars', 'float32', 'deg'),
             'adc': ([8], 'sys', 'uint8', ''),
             'error_status': ([], 'attrs', 'float32', ''),
             'pressure': ([], 'data_vars', 'float32', 'dbar'),
             'pressure_std': ([], 'data_vars', 'float32', 'dbar'),
             'vel': (['nc', 4], 'data_vars', 'float32', 'm/s'),
             'amp': (['nc', 4], 'data_vars', 'uint8', 'counts'),
             'corr': (['nc', 4], 'data_vars', 'uint8', 'counts'),
             'prcnt_gd': (['nc', 4], 'data_vars', 'uint8', '%'),
             'status': (['nc', 4], 'data_vars', 'float32', ''),
             'dist_bt': ([4], 'data_vars', 'float32', 'm'),
             'vel_bt': ([4], 'data_vars', 'float32', 'm/s'),
             'corr_bt': ([4], 'data_vars', 'uint8', 'counts'),
             'amp_bt': ([4], 'data_vars', 'uint8', 'counts'),
             'prcnt_gd_bt': ([4], 'data_vars', 'uint8', '%'),
             'time': ([], 'coords', 'float64', ''),
             'alt_dist': ([], 'data_vars', 'float32', 'm'),
             'alt_rssi': ([], 'data_vars', 'uint8', 'dB'),
             'alt_eval': ([], 'data_vars', 'uint8', 'dB'),
             'alt_status': ([], 'data_vars', 'uint8', 'bit'),
             'time_gps': ([], 'coords', 'float64', ''),
             'clock_offset_UTC_gps': ([], 'data_vars', 'float64', 's'),
             'latitude_gps': ([], 'data_vars', 'float32', 'deg'),
             'longitude_gps': ([], 'data_vars', 'float32', 'deg'),
             'avg_speed_gps': ([], 'data_vars', 'float32', 'm/s'),
             'avg_dir_gps': ([], 'data_vars', 'float32', 'deg'),
             'speed_made_good_gps': ([], 'data_vars', 'float32', 'm/s'),
             'dir_made_good_gps': ([], 'data_vars', 'float32', 'deg'),
             'flags_gps': ([], 'data_vars', 'float32', 'bits'),
             'fix_gps': ([], 'data_vars', 'int8', '1'),
             'n_sat_gps': ([], 'data_vars', 'int8', 'count'),
             'hdop_gps': ([], 'data_vars', 'float32', '1'),
             'elevation_gps': ([], 'data_vars', 'float32', 'm'),
             'rtk_age_gps': ([], 'data_vars', 'float32', 's'),
             'speed_over_grnd_gps': ([], 'data_vars', 'float32', 'm/s'),
             'dir_over_grnd_gps': ([], 'data_vars', 'float32', 'deg'),
             'heading_gps': ([], 'data_vars', 'float64', 'deg'),
             'dist_nmea': ([], 'data_vars', 'float32', 'm'),
             'vel_sl': (['nc', 4], 'data_vars', 'float32', 'm/s'),
             'corr_sl': (['nc', 4], 'data_vars', 'uint8', 'counts'),
             'amp_sl': (['nc', 4], 'data_vars', 'uint8', 'counts'),
             'prcnt_gd_sl': (['nc', 4], 'data_vars', 'uint8', '%'),
             'status_sl': (['nc', 4], 'data_vars', 'float32', ''),
             }


def _get(dat, nm):
    grp = data_defs[nm][1]
    if grp is None:
        return dat[nm]
    else:
        return dat[grp][nm]


def _in_group(dat, nm):
    grp = data_defs[nm][1]
    if grp is None:
        return nm in dat
    else:
        return nm in dat[grp]


def _pop(dat, nm):
    grp = data_defs[nm][1]
    if grp is None:
        dat.pop(nm)
    else:
        dat[grp].pop(nm)


def _setd(dat, nm, val):
    grp = data_defs[nm][1]
    if grp is None:
        dat[nm] = val
    else:
        dat[grp][nm] = val


def _idata(dat, nm, sz):
    group = data_defs[nm][1]
    dtype = data_defs[nm][2]
    units = data_defs[nm][3]
    arr = np.empty(sz, dtype=dtype)
    if dtype.startswith('float'):
        arr[:] = np.NaN
    dat[group][nm] = arr
    dat['units'][nm] = units
    return dat


def _get_size(name, n=None, ncell=0):
    sz = list(data_defs[name][0])  # create a copy!
    if 'nc' in sz:
        sz.insert(sz.index('nc'), ncell)
        sz.remove('nc')
    if n is None:
        return tuple(sz)
    return tuple(sz + [n])


class _variable_setlist(set):
    def __iadd__(self, vals):
        if vals[0] not in self:
            self |= set(vals)
        return self


class _ensemble():
    n_avg = 1
    k = -1  # This is the counter for filling the ensemble object

    def __getitem__(self, nm):
        return getattr(self, nm)

    def __init__(self, navg, n_cells):
        if navg is None or navg == 0:
            navg = 1
        self.n_avg = navg
        self.n_cells = n_cells
        for nm in data_defs:
            setattr(self, nm,
                    np.zeros(_get_size(nm, n=navg, ncell=n_cells),
                             dtype=data_defs[nm][2]))

    def clean_data(self,):
        self['vel'][self['vel'] == -32.768] = np.NaN



import numpy as np
from struct import unpack
from os.path import expanduser


class bin_reader():
    """
    Reads binary data files. It is mostly for development purposes, to
    simplify learning a data file's format. Reading binary data files should
    minimize the number of calls to struct.unpack and file.read because many
    calls to these functions (i.e. using the code in this module) are slow.
    """
    _size_factor = {'B': 1, 'b': 1, 'H': 2,
                    'h': 2, 'L': 4, 'l': 4, 'f': 4, 'd': 8}
    _frmt = {np.uint8: 'B', np.int8: 'b',
             np.uint16: 'H', np.int16: 'h',
             np.uint32: 'L', np.int32: 'l',
             float: 'f', np.float32: 'f',
             np.double: 'd', np.float64: 'd',
             }

    @property
    def pos(self,):
        return self.f.tell()

    def __enter__(self,):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __init__(self, fname, endian='<', checksum_size=None, debug_level=0):
        """
        Default to little-endian '<'...
        *checksum_size* is in bytes, if it is None or False, this
         function does not perform checksums.
        """
        self.endian = endian
        self.f = open(expanduser(fname), 'rb')
        self.f.seek(0, 2)
        self.fsize = self.tell()
        self.f.seek(0, 0)
        self.close = self.f.close
        if checksum_size:
            pass
            #self.cs = checksum(self, 0, checksum_size)
        else:
            self.cs = checksum_size
        self.debug_level = debug_level

    # def checksum(self,):
    #     """
    #     The next byte(s) are the expected checksum.  Perform the checksum.
    #     """
    #     if self.cs:
    #         cs = self.read(1, self.cs._frmt)
    #         self.cs(cs, True)
    #     else:
    #         raise Exception('CheckSum not requested for this file')

    def tell(self,):
        return self.f.tell()

    def seek(self, pos, rel=1):
        return self.f.seek(pos, rel)

    def reads(self, n):
        """
        Read a string of n characters.
        """
        val = self.f.read(n)
        self.cs and self.cs.add(val)
        try:
            val = val.decode('utf-8')
        except:
            if self.debug_level > 5:
                print("ERROR DECODING: {}".format(val))
            pass
        return val

    def read(self, n, frmt):
        val = self.f.read(n * self._size_factor[frmt])
        if not val:  # If val is empty we are at the end of the file.
            return None
        self.cs and self.cs.add(val)
        if n == 1:
            return unpack(self.endian + frmt * n, val)[0]
        else:
            return np.array(unpack(self.endian + frmt * n, val))

    def read_ui8(self, n):
        return self.read(n, 'B')

    def read_float(self, n):
        return self.read(n, 'f')

    def read_double(self, n):
        return self.read(n, 'd')

    read_f32 = read_float
    read_f64 = read_double

    def read_i8(self, n):
        return self.read(n, 'b')

    def read_ui16(self, n):
        return self.read(n, 'H')

    def read_i16(self, n):
        return self.read(n, 'h')

    def read_ui32(self, n):
        return self.read(n, 'L')

    def read_i32(self, n):
        return self.read(n, 'l')

#ics = 0  # This is a holder for the checksum index
# class checksum():
#     # Checksum for TRDI
#     def __init__(self, file, val, size, error_behavior='exception'):
#         """
#         Value *val* to initialize the checksum with,
#         and the *size* of the checksum (in bytes, currently this can only be 1,2,4 or 8).
#         """
#         self.file = file
#         self.init(val, size, error_behavior)

#     def init(self, val, size, error_behavior='exception'):
#         self._cs = val
#         self._size = size
#         self._rem = ''
#         self._frmt = {1: 'B', 2: 'H', 4: 'L', 8: 'Q'}[size]
#         self._mask = (2 ** (8 * size) - 1)
#         ## I'd like to add different behavior for the checksum
#         self._error_behavior = error_behavior

#     def error(self, val):
#         if val == 'rem':
#             message = 'A remainder exists in the checksum.'
#         else:
#             message = 'Checksum failed at %d, with a difference of %d.' % (
#                 self.file.tell(), val)
#         if self._error_behavior == 'warning':
#             print('Warning: ' + message)
#         elif self._error_behavior == 'silent':
#             pass
#         else:
#             raise Exception(message)

#     def __call__(self, val, remove_val=False):
#         """
#         Compare the checksum to *val*.
#         *remove_val* specifies whether *val* should be removed from
#           self._cs because it was already added to it.
#         """
#         if self._rem:
#             self.error('rem')
#         retval = (self._cs - val * remove_val & self._mask) - val
#         if retval:
#             self.error(retval)
#         return retval  # returns zero if there is no remainder

#     def add(self, valcs):
#         """
#         Add the data in *valcs* to the checksum.
#         """
#         if self._rem:  # If the cs remainder is not empty:
#             lr = self._rem.__len__()
#             ics = self._size - lr
#             self._rem, valcs = self._rem + valcs[:ics], valcs[ics:]
#             if lr == self._size:
#                 self._cs += unpack(self.file.endian + self._frmt, self._rem)[0]
#                 self._rem = ''
#         if valcs:  # If valcs is not empty:
#             ics = (valcs.__len__() / self._size) * self._size
#             for v in unpack(self.file.endian + self._frmt * (ics / self._size), valcs[:ics]):
#                 self._cs += v
#             self._rem += valcs[ics:]

#     __iadd__ = add



import numpy as np
import warnings
from . import base as rotb


def _beam2inst(dat, reverse=False, force=False):
    # Order of rotations matters
    # beam->head(ADV instrument head)->inst(ADV battery case|imu)
    if reverse:
        # First rotate velocities from ADV inst frame back to head frame
        dat = _rotate_inst2head(dat, reverse=reverse)
        # Now rotate from the head frame to the beam frame
        dat = rotb._beam2inst(dat, reverse=reverse, force=force)

    # inst(ADV battery case|imu)->head(ADV instrument head)->beam
    else:
        # First rotate velocities from beam to ADV head frame
        dat = rotb._beam2inst(dat, force=force)
        # Then rotate from ADV head frame to ADV inst frame
        dat = _rotate_inst2head(dat)

    # Set the docstring to match the default rotation func
    _beam2inst.__doc_ = rotb._beam2inst.__doc__

    return dat


def _inst2earth(advo, reverse=False, rotate_vars=None, force=False):
    """
    Rotate data in an ADV object to the earth from the instrument
    frame (or vice-versa).

    Parameters
    ----------
    advo : The adv object containing the data.

    reverse : bool (default: False)
           If True, this function performs the inverse rotation
           (earth->inst).

    rotate_vars : iterable
      The list of variables to rotate. By default this is taken from
      advo.props['rotate_vars'].

    force : Do not check which frame the data is in prior to
      performing this rotation.

    """
    if reverse:  # earth->inst
        # The transpose of the rotation matrix gives the inverse
        # rotation, so we simply reverse the order of the einsum:
        sumstr = 'jik,j...k->i...k'
        cs_now = 'earth'
        cs_new = 'inst'
    else:  # inst->earth
        sumstr = 'ijk,j...k->i...k'
        cs_now = 'inst'
        cs_new = 'earth'

    if rotate_vars is None:
        if 'rotate_vars' in advo.attrs:
            rotate_vars = advo.rotate_vars
        else:
            rotate_vars = ['vel']

    cs = advo.coord_sys.lower()
    if not force:
        if cs == cs_new:
            print("Data is already in the '%s' coordinate system" % cs_new)
            return
        elif cs != cs_now:
            raise ValueError(
                "Data must be in the '%s' frame when using this function" %
                cs_now)

    if hasattr(advo, 'orientmat'):
        omat = advo['orientmat'].values
    else:
        if 'vector' in advo.inst_model.lower():
            orientation_down = advo['orientation_down']

        omat = _calc_omat(advo['heading'].values, advo['pitch'].values,
                          advo['roll'].values, orientation_down)

    # Take the transpose of the orientation to get the inst->earth rotation
    # matrix.
    rmat = np.rollaxis(omat, 1)

    _dcheck = rotb._check_rotmat_det(rmat)
    if not _dcheck.all():
        warnings.warn("Invalid orientation matrix (determinant != 1) at indices: {}. "
                      "If rotated, data at these indices will be erroneous."
                      .format(np.nonzero(~_dcheck)[0]), UserWarning)

    for nm in rotate_vars:
        n = advo[nm].shape[0]
        if n != 3:
            raise Exception("The entry {} is not a vector, it cannot "
                            "be rotated.".format(nm))
        advo[nm].values = np.einsum(sumstr, rmat, advo[nm])

    advo = rotb._set_coords(advo, cs_new)

    return advo


def _calc_omat(hh, pp, rr, orientation_down=None):
    rr = rr.copy()
    pp = pp.copy()
    hh = hh.copy()
    if np.isnan(rr[-1]) and np.isnan(pp[-1]) and np.isnan(hh[-1]):
        # The end of the data may not have valid orientations
        lastgd = np.nonzero(~np.isnan(rr + pp + hh))[0][-1]
        rr[lastgd:] = rr[lastgd]
        pp[lastgd:] = pp[lastgd]
        hh[lastgd:] = hh[lastgd]
    if orientation_down is not None:
        # For Nortek Vector ADVs: 'down' configuration means the head was
        # pointing 'up', where the 'up' orientation corresponds to the
        # communication cable being up.  Check the Nortek coordinate
        # transform matlab script for more info.
        rr[orientation_down.astype(bool)] += 180

    return _euler2orient(hh, pp, rr)


def _rotate_inst2head(advo, reverse=False):
    if not _check_inst2head_rotmat(advo):
        # This object doesn't have a head2inst_rotmat, so we do nothing.
        return advo
    if reverse:  # head->inst
        advo['vel'].values = np.dot(advo['inst2head_rotmat'].T, advo['vel'])
    else:  # inst->head
        advo['vel'].values = np.dot(advo['inst2head_rotmat'], advo['vel'])

    return advo


def _check_inst2head_rotmat(advo):
    if advo.get('inst2head_rotmat', None) is None:
        # This is the default value, and we do nothing.
        return False
    if not advo.inst2head_rotmat_was_set:
        raise Exception("The inst2head rotation matrix exists in props, "
                        "but it was not set using `set_inst2head_rotmat.")
    if not rotb._check_rotmat_det(advo.inst2head_rotmat.values):
        raise ValueError("Invalid inst2head_rotmat (determinant != 1).")
    return True


def _earth2principal(advo, reverse=False):
    """
    Rotate data in an ADV dataset to/from principal axes. Principal
    heading must be within the dataset.

    All data in the advo.attrs['rotate_vars'] list will be
    rotated by the principal heading, and also if the data objet has an
    orientation matrix (orientmat) it will be rotated so that it
    represents the orientation of the ADV in the principal
    (reverse:earth) frame.

    Parameters
    ----------
    advo : The adv object containing the data.
    reverse : bool (default: False)
           If True, this function performs the inverse rotation
           (principal->earth).

    """
    # This is in degrees CW from North
    ang = np.deg2rad(90 - advo.principal_heading)
    # convert this to radians CCW from east (which is expected by
    # the rest of the function)

    if reverse:
        cs_now = 'principal'
        cs_new = 'earth'
    else:
        ang *= -1
        cs_now = 'earth'
        cs_new = 'principal'

    cs = advo.coord_sys.lower()
    if cs == cs_new:
        print('Data is already in the %s coordinate system' % cs_new)
        return
    elif cs != cs_now:
        raise ValueError(
            'Data must be in the {} frame '
            'to use this function'.format(cs_now))

    # Calculate the rotation matrix:
    cp, sp = np.cos(ang), np.sin(ang)
    rotmat = np.array([[cp, -sp, 0],
                       [sp, cp, 0],
                       [0, 0, 1]], dtype=np.float32)

    # Perform the rotation:
    for nm in advo.rotate_vars:
        dat = advo[nm].values
        dat[:2] = np.einsum('ij,j...->i...', rotmat[:2, :2], dat[:2])
        advo[nm].values = dat.copy()

    # Finalize the output.
    advo = rotb._set_coords(advo, cs_new)

    return advo


def _euler2orient(heading, pitch, roll, units='degrees'):
    # For Nortek data only.
    # The heading, pitch, roll used here are from the Nortek binary files.

    # Heading input is clockwise from North
    # Returns a rotation matrix that rotates earth (ENU) -> inst.
    # This is based on the Nortek `Transforms.m` file, available in
    # the refs folder.
    if units.lower() == 'degrees':
        pitch = np.deg2rad(pitch)
        roll = np.deg2rad(roll)
        heading = np.deg2rad(heading)

    # The definition of heading below is consistent with the right-hand-rule;
    # heading is the angle positive counterclockwise from North of the y-axis.

    # This also involved swapping the sign on sh in the def of omat
    # below from the values provided in the Nortek Matlab script.
    heading = (np.pi / 2 - heading)

    ch = np.cos(heading)
    sh = np.sin(heading)
    cp = np.cos(pitch)
    sp = np.sin(pitch)
    cr = np.cos(roll)
    sr = np.sin(roll)

    # Note that I've transposed these values (from what is defined in
    # Nortek matlab script), so that the omat is earth->inst
    omat = np.empty((3, 3, len(sh)), dtype=np.float32)
    omat[0, 0, :] = ch * cp
    omat[1, 0, :] = -ch * sp * sr - sh * cr
    omat[2, 0, :] = -ch * cr * sp + sh * sr
    omat[0, 1, :] = sh * cp
    omat[1, 1, :] = -sh * sp * sr + ch * cr
    omat[2, 1, :] = -sh * cr * sp - ch * sr
    omat[0, 2, :] = sp
    omat[1, 2, :] = sr * cp
    omat[2, 2, :] = cp * cr

    return omat
